(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
header {* Term *}
theory Term imports AbstractRewriting Position Show Util begin
text {* A theory on first-order terms. *}

text {* A \emph{signature} is a set of function symbols,each associated with an \emph{arity}, 
denoting the number of arguments it expects. *}
types 'f sig = "('f set \<times> ('f \<Rightarrow> nat))"

text {* The @{text arity} of a function symbol is the number of arguments it takes. *}
abbreviation arity :: "'f sig \<Rightarrow> 'f \<Rightarrow> nat" where "arity \<Sigma> f \<equiv> (snd \<Sigma>) f"

text {* For convenience @{text "f :s \<Sigma>"} can be written instead of @{text "f \<in> (fst F)"}, i.e.,
to check whether the functionsymbol~@{text f} is contained in the signature~@{text \<Sigma>}. *}
(* The mixfix declaration uses the same priorities as for set membership. *)
abbreviation sig_in :: "'f \<Rightarrow> 'f sig \<Rightarrow> bool" ("(_/ :s _)" [50, 51] 50)
where "f :s F \<equiv> f \<in> (fst F)"

text {* (First-order) \emph{terms} are built from function symbols and variables. *}
datatype ('f,'v)"term" = Var 'v | Fun 'f "('f,'v)term list"

text {* Resolve problems with recurisve dependencies in generated code. *}
lemma [code]:
  shows "HOL.eq (Var x) (Var y) \<longleftrightarrow> HOL.eq x y"
  and "HOL.eq (Fun f ts) (Fun g us) \<longleftrightarrow> HOL.eq f g \<and> list_all2 HOL.eq ts us"
  by (simp_all add: eq list_all2_eq [symmetric])

text {* An alternative induction scheme for terms. *}
lemma
 fixes P :: "('f,'v)term \<Rightarrow> bool"
 assumes "\<And>x. P(Var x)" and "\<And>f ss. (\<And>s. s \<in> set ss \<Longrightarrow> P s) \<Longrightarrow> P(Fun f ss)"
 shows term_induct[case_names Var Fun, induct type: "term"]: "P t" and "\<And>t. t \<in> set ts \<Longrightarrow> P t"
using assms by (induct t and ts) auto

text {* Auxiliary lemmas for structural decreasingness. *}
(* They are used in the termination proofs of some of the `fun' definitions. *)
lemma [simp]: "t \<in> set ts \<Longrightarrow> size t < Suc(list_size size ts)" by (induct ts) auto
lemma [simp]: "t \<in> set ts \<Longrightarrow> size t < Suc(Suc(size s + list_size size ts))" by (induct ts) auto

text {* A term is a \emph{well-formed} (wf)---with respect to a given signature---if every 
occurrence of a function symbol is used with a number of arguments corresponding to its arity. *}
fun wf_term :: "'f sig \<Rightarrow> ('f,'v)term \<Rightarrow> bool"
where "wf_term \<Sigma> (Var x) = True"
    | "wf_term \<Sigma> (Fun f ss) = (f :s \<Sigma> \<and> arity \<Sigma> f = length ss \<and> (\<forall>s \<in> set ss. wf_term \<Sigma> s))"


subsection {* shows Instance for Terms *}

instantiation "term" :: ("shows","shows")"shows" begin
  fun shows_term :: "('f::shows,'v::shows)term \<Rightarrow> shows"
  where "shows_term(Var x) = shows x"
    | "shows_term(Fun f ts) = (
      shows f +@+ shows_list_gen id [] ''('' '','' '')'' (map shows_term ts)
    )"
  definition "shows_list_term(is::('f::shows,'v::shows)term list) \<equiv> shows_list_aux shows is"

  lemma assoc_term: "shows(t::('f::shows,'v::shows)term) r @ s = shows t (r@s)"
  proof (induct t arbitrary: r s)
    case (Var x) show ?case by simp
  next
    case (Fun f ts)
    hence "\<And>r s. \<forall>t\<in>set ts. id (shows t) r @ s = id (shows t) (r @ s)" by auto
    hence elt: "\<And>r s. \<forall>t\<in>set(map shows ts). id t r @ s = id t (r @ s)" by auto
    show ?case unfolding shows_term.simps
      unfolding o_def unfolding assoc
      unfolding shows_list_gen_assoc[OF elt,of "[]" "''(''" "'',''" "'')''" r s] by simp
  qed
  lemma assoc_term_elt: "\<forall>(t::('f::shows,'v::shows)term)\<in>set ts. shows t r @ s = shows t (r@s)"
    using assoc_term by auto
 instance by
    (intro_classes,rule assoc_term,unfold shows_list_term_def)
    (rule shows_list_aux_assoc[OF assoc_term_elt])
end

subsection {* Variables *}
primrec is_var :: "('f,'v)term \<Rightarrow> bool" 
where "is_var(Var _) = True" 
    | "is_var(Fun _ _) = False"

lemma is_var_imp_Var[elim]: "is_var t \<Longrightarrow> (\<And>x. t = Var x \<Longrightarrow> P) \<Longrightarrow> P" by (cases t) simp_all
lemma not_is_var_imp_Fun[elim]: "\<not>(is_var t) \<Longrightarrow> (\<And>f ts. t = Fun f ts \<Longrightarrow> P) \<Longrightarrow> P"
by (cases t) auto

fun the_var :: "('f,'v)term \<Rightarrow> 'v"
where "the_var(Var x) = x"

text {* The set of all variables occuring in a term. *}
fun vars_term :: "('f,'v)term \<Rightarrow> 'v set"
where "vars_term(Var x) = {x}"
    | "vars_term(Fun _ ts) = \<Union>set(map vars_term ts)"

text {* The list of variables in a term (without removing duplicates). *}
fun vars_term_list' :: "('f,'v)term \<Rightarrow> 'v list"
where "vars_term_list'(Var x) = [x]"
    | "vars_term_list'(Fun _ ts) = flat_map vars_term_list' ts"

lemma vars_term_list'_sound[simp]:
  fixes t::"('f,'v)term" shows "set(vars_term_list' t) = vars_term t" by (induct t) simp_all
declare vars_term_list'.simps[simp del]

fun vars_term_list :: "('f,'v)term \<Rightarrow> 'v list"
where "vars_term_list t = remdups(vars_term_list' t)"

lemma vars_term_list_sound[simp]:
  fixes t::"('f,'v)term" shows "set(vars_term_list t) = vars_term t" by simp
declare vars_term_list.simps[simp del]


text {* The set of all function symbols occuring in a term. *}
fun funs_term :: "('f,'v)term \<Rightarrow> 'f set"
where "funs_term(Var _) = {}"
    | "funs_term(Fun f ts) = {f} \<union> \<Union>set(map funs_term ts)"

text {* The list of function symbols in a term (without removing duplicates). *}
fun funs_term_list' :: "('f,'v)term \<Rightarrow> 'f list"
where "funs_term_list'(Var _) = []"
    | "funs_term_list'(Fun f ts) = f # flat_map funs_term_list' ts"

lemma funs_term_list'_sound[simp]:
  fixes t::"('f,'v)term" shows "set(funs_term_list' t) = funs_term t" by (induct t) simp_all
declare funs_term_list'.simps[simp del]

fun funs_term_list :: "('f,'v)term \<Rightarrow> 'f list"
where "funs_term_list t = remdups(funs_term_list' t)"

lemma funs_term_list_sound[simp]:
  fixes t::"('f,'v)term" shows "set(funs_term_list t) = funs_term t" by simp
declare funs_term_list.simps[simp del]

text {* The set of all function symbol and arity pairs occuring in a term. *}
fun funas_term :: "('f,'v)term \<Rightarrow> ('f\<times>nat)set"
where "funas_term(Var _) = {}"
    | "funas_term(Fun f ts) = {(f,length ts)} \<union> \<Union>set(map funas_term ts)"

text {* The list of function symbol and arity pairs in a term (without removing duplicates). *}
fun funas_term_list' :: "('f,'v)term \<Rightarrow> ('f\<times>nat)list"
where "funas_term_list'(Var _) = []"
    | "funas_term_list'(Fun f ts) = (f,length ts) # flat_map funas_term_list' ts"

lemma funas_term_list'_sound[simp]:
  fixes t::"('f,'v)term" shows "set(funas_term_list' t) = funas_term t" by (induct t) simp_all
declare funas_term_list'.simps[simp del]

fun funas_term_list :: "('f,'v)term \<Rightarrow> ('f\<times>nat)list"
where "funas_term_list t = remdups(funas_term_list' t)"

lemma funas_term_list_sound[simp]:
  fixes t::"('f,'v)term" shows "set(funas_term_list t) = funas_term t" by simp
declare funas_term_list.simps[simp del]

text {* A term is called \emph{ground} if it does not contain
any variables. *}
fun ground :: "('f,'v)term \<Rightarrow> bool"
where "ground(Var x) = False"
    | "ground(Fun f ts) = (\<forall>t \<in> set ts. ground t)"

text {* The \emph{root symbol} of a term is defined by: *}
fun root :: "('f,'v)term \<Rightarrow> 'f option"
where "root(Var x) = None"
    | "root(Fun f ts) = Some f"

text {* The \emph{size} of a term is the number of variables and function symbols occurring in it. *}
fun sz :: "('f,'v)term \<Rightarrow> nat" ("|_|")
where "sz(Var x) = 1"
    | "sz(Fun f ts) = 1 + listsum(map sz ts)"

text {* The number of function symbols occuring in a term. *}
fun count_fs :: "('f,'v)term \<Rightarrow> nat" ("\<parallel>_\<parallel>")
where "\<parallel>Var x\<parallel> = 0"
    | "\<parallel>Fun f ts\<parallel> = 1 + listsum(map (\<lambda>t. \<parallel>t\<parallel>) ts)"

text {* The \emph{depth} of a term is defined as follows: *}
fun depth :: "('f,'v)term \<Rightarrow> nat"
where "depth(Var x) = 0"
    | "depth(Fun f []) = 0"
    | "depth(Fun f ts) = 1 + Max(set(map depth ts))"


subsection {* Subterms *}

text {* The \emph{superterm} relation. *}
inductive_set supteq :: "(('f,'v)term \<times> ('f,'v)term)set"
where refl[intro]: "(t,t) \<in> supteq"
    | subteq_of_arg[intro]: "u \<in> set ss \<Longrightarrow> (u,t) \<in> supteq \<Longrightarrow> (Fun f ss,t) \<in> supteq"

text {* The \emph{proper superterm} relation. *}
inductive_set supt :: "(('f,'v)term \<times> ('f,'v)term)set"
where arg[intro]: "s \<in> set ss \<Longrightarrow> (Fun f ss,s) \<in> supt"
    | subt_of_arg[intro]: "s \<in> set ss \<Longrightarrow> (s,t) \<in> supt \<Longrightarrow> (Fun f ss,t) \<in> supt"

subsubsection {* Syntax for Subterms *}
text {* Infix syntax. *}
notation
  supteq ("op |>=") and
  supteqp ("(_/ |>= _)" [51, 51] 50) and
  supt ("op |>") and
  suptp ("(_/ |> _)" [51, 51] 50)

notation (xsymbols)
  supteq ("op \<unrhd>") and
  supteqp ("(_/ \<unrhd> _)" [51, 51] 50) and
  supt ("op \<rhd>") and
  suptp ("(_/ \<rhd> _)" [51, 51] 50)

abbreviation (input) subteq (infix "<|=" 50) where "s <|= t \<equiv> t |>= s"
notation (input) subteq (infix "\<unlhd>" 50)

abbreviation (input) subt (infix "<|" 50) where "s <| t \<equiv> t |> s"
notation (input) subt (infix "\<lhd>" 50)

text {* Quantifier syntax. *}

syntax
  "_All_supteq" :: "[idt, 'a, bool] \<Rightarrow> bool" ("(3ALL _|>=_./ _)" [0, 0, 10] 10)
  "_Ex_supteq" :: "[idt, 'a, bool] \<Rightarrow> bool" ("(3EX _|>=_./ _)" [0, 0, 10] 10)
  "_All_supt" :: "[idt, 'a, bool] \<Rightarrow> bool" ("(3ALL _|>_./ _)" [0, 0, 10] 10)
  "_Ex_supt" :: "[idt, 'a, bool] \<Rightarrow> bool" ("(3EX _|>_./ _)" [0, 0, 10] 10)

  "_All_subteq" :: "[idt, 'a, bool] \<Rightarrow> bool" ("(3ALL _<|=_./ _)" [0, 0, 10] 10)
  "_Ex_subteq" :: "[idt, 'a, bool] \<Rightarrow> bool" ("(3EX _<|=_./ _)" [0, 0, 10] 10)
  "_All_subt" :: "[idt, 'a, bool] \<Rightarrow> bool" ("(3ALL _<|_./ _)" [0, 0, 10] 10)
  "_Ex_subt" :: "[idt, 'a, bool] \<Rightarrow> bool" ("(3EX _<|_./ _)" [0, 0, 10] 10)

syntax (xsymbols)
  "_All_supteq" :: "[idt, 'a, bool] \<Rightarrow> bool" ("(3\<forall>_\<unrhd>_./ _)" [0, 0, 10] 10)
  "_Ex_supteq" :: "[idt, 'a, bool] \<Rightarrow> bool" ("(3\<exists>_\<unrhd>_./ _)" [0, 0, 10] 10)
  "_All_supt" :: "[idt, 'a, bool] \<Rightarrow> bool" ("(3\<forall>_\<rhd>_./ _)" [0, 0, 10] 10)
  "_Ex_supt" :: "[idt, 'a, bool] \<Rightarrow> bool" ("(3\<exists>_\<rhd>_./ _)" [0, 0, 10] 10)

  "_All_subteq" :: "[idt, 'a, bool] \<Rightarrow> bool" ("(3\<forall>_\<unlhd>_./ _)" [0, 0, 10] 10)
  "_Ex_subteq" :: "[idt, 'a, bool] \<Rightarrow> bool" ("(3\<exists>_\<unlhd>_./ _)" [0, 0, 10] 10)
  "_All_subt" :: "[idt, 'a, bool] \<Rightarrow> bool" ("(3\<forall>_\<lhd>_./ _)" [0, 0, 10] 10)
  "_Ex_subt" :: "[idt, 'a, bool] \<Rightarrow> bool" ("(3\<exists>_\<lhd>_./ _)" [0, 0, 10] 10)

(* for parsing *)
translations
  "ALL x|>=y. P" => "ALL x. x |>= y \<longrightarrow> P"
  "EX x|>=y. P"  => "EX x. x |>= y \<and> P"
  "ALL x|>y. P"  => "ALL x. x |> y \<longrightarrow> P"
  "EX x|>y. P"   => "EX x. x |> y \<and> P"
  
  "ALL x<|=y. P" => "ALL x. x <|= y \<longrightarrow> P"
  "EX x<|=y. P"  => "EX x. x <|= y \<and> P"
  "ALL x<|y. P"  => "ALL x. x <| y \<longrightarrow> P"
  "EX x<|y. P"   => "EX x. x <| y \<and> P"

(* for printing *)
print_translation {*
let
  val All_binder = Syntax.binder_name @{const_syntax All};
  val Ex_binder = Syntax.binder_name @{const_syntax Ex};
  val impl = @{const_syntax "op -->"};
  val conj = @{const_syntax "op &"};
  val supt = @{const_syntax suptp};
  val supteq = @{const_syntax supteqp};

  val trans =
   [((All_binder, impl, supt), ("_All_supt", "_All_subt")),
    ((All_binder, impl, supteq), ("_All_supteq", "_All_subteq")),
    ((Ex_binder, conj, supt), ("_Ex_supt", "_Ex_subt")),
    ((Ex_binder, conj, supteq), ("_Ex_supteq", "_Ex_subteq"))];

  fun matches_bound v t = 
     case t of (Const ("_bound", _) $ Free (v', _)) => (v = v')
              | _ => false
  fun contains_var v = Term.exists_subterm (fn Free (x, _) => x = v | _ => false)
  fun mk v c n P = Syntax.const c $ Syntax.mark_bound v $ n $ P

  fun tr' q = (q,
    fn [Const ("_bound", _) $ Free (v, _), Const (c, _) $ (Const (d, _) $ t $ u) $ P] =>
      (case AList.lookup (op =) trans (q, c, d) of
        NONE => raise Match
      | SOME (l, g) =>
          if matches_bound v t andalso not (contains_var v u) then mk v l u P
          else if matches_bound v u andalso not (contains_var v t) then mk v g t P
          else raise Match)
     | _ => raise Match);
in [tr' All_binder, tr' Ex_binder] end
*}

subsubsection {* Transitivity Reasoning for Subterms *}
lemma suptp_trans[trans]: "s \<rhd> t \<Longrightarrow> t \<rhd> u \<Longrightarrow> s \<rhd> u"
by (induct s t rule: suptp.induct) auto
lemmas supt_trans[trans] = suptp_trans[to_set]

lemma supteqp_trans[trans]: "s \<unrhd> t \<Longrightarrow> t \<unrhd> u \<Longrightarrow> s \<unrhd> u"
by (induct s t rule: supteqp.induct) auto
lemmas supteq_trans[trans] = supteqp_trans[to_set]

lemma [simp]: "s \<in> set ss \<Longrightarrow> s \<rhd> t \<Longrightarrow> size t < size s \<Longrightarrow> size t < Suc(list_size size ss)"
by (induct ss) auto

lemma [simp]:
  "s \<in> set ss \<Longrightarrow> (s,t) \<in> supt \<Longrightarrow> size t < size s \<Longrightarrow> size t < Suc(list_size size ss)"
by (induct ss) auto

lemma suptp_size: "s \<rhd> t \<Longrightarrow> size s > size t" by (induct rule: suptp.induct) auto
lemmas supt_size = suptp_size[to_set]

lemma suptp_imp_neq: "s \<rhd> t \<Longrightarrow> s \<noteq> t" using suptp_size by fast
lemmas supt_imp_neq = suptp_imp_neq[to_set]

lemma supteqp_var: "x \<in> vars_term t \<Longrightarrow> t \<unrhd> Var x"
proof (induct t)
  case (Var y) thus ?case by (cases "x = y", auto)
next
  case (Fun f ss) thus ?case by auto
qed


lemma term_not_arg[simp]: "Fun f ss \<notin> set ss" (is "?t \<notin> set ss")
proof
 assume "?t \<in> set ss"
 hence "?t \<rhd> ?t" by (rule suptp.arg)
 hence "?t \<noteq> ?t" by (rule suptp_imp_neq)
 thus False by simp
qed

lemma arg_not_suptp: assumes "s \<in> set ss" and "s \<rhd> Fun f ss" (is "s \<rhd> ?t") shows " False"
proof -
 from `s \<in> set ss` have "?t \<rhd> s" by auto
 hence "size ?t > size s" by (rule suptp_size)
 from `s \<rhd> ?t` have "size s > size ?t" by (rule suptp_size)
 with `size ?t > size s` show False by simp
qed

lemma suptp_iff_supteqp_neq[iff]: "(s \<rhd> t) = (s \<unrhd> t \<and> s \<noteq> t)"
proof
 assume "s \<rhd> t" thus "s \<unrhd> t \<and> s \<noteq> t"
 proof (induct rule: suptp.induct)
  case (arg t ss f) show ?case
  proof
   from arg show "Fun f ss \<unrhd> t" by auto
  next
   from arg have "size t < size (Fun f ss)" by auto
   thus "Fun f ss \<noteq> t" by auto
  qed
 next
  case (subt_of_arg s ss t f)
  have "s \<unrhd> s" ..
  from `s \<in> set ss` have "Fun f ss \<unrhd> s" by auto
  from `s \<unrhd> t \<and> s \<noteq> t` have "s \<unrhd> t" ..
  with `Fun f ss \<unrhd> s` have first: "Fun f ss \<unrhd> t" by (rule supteqp_trans)
  from `s \<in> set ss` and `s \<rhd> t` have "Fun f ss \<rhd> t" ..
  hence second: "Fun f ss \<noteq> t" by (rule suptp_imp_neq)
  from first and second show "Fun f ss \<unrhd> t \<and> Fun f ss \<noteq> t" by auto
 qed
next
 assume "s \<unrhd> t \<and> s \<noteq> t"
 hence "s \<unrhd> t" and "s \<noteq> t" by auto
 thus "s \<rhd> t" by (induct rule: supteqp.induct) auto
qed

lemmas supt_iff_supteq_neq[iff] = suptp_iff_supteqp_neq[to_set]

lemma suptp_asym: "s \<rhd> t \<Longrightarrow> \<not> (t \<rhd> s)"
proof
 assume "s \<rhd> t" and "t \<rhd> s" hence "s \<rhd> s" by (rule suptp_trans)
 thus False unfolding suptp_iff_supteqp_neq by simp
qed

lemma supteqp_iff_suptp_or_eq: "(s \<unrhd> t) = (s \<rhd> t \<or> s = t)" using suptp_iff_supteqp_neq by auto

lemma supteqp_cases[consumes 1,elim]: "\<lbrakk>s \<unrhd> t; s = t \<Longrightarrow> P; s \<rhd> t \<Longrightarrow> P\<rbrakk> \<Longrightarrow> P"
unfolding supteqp_iff_suptp_or_eq by auto

lemmas supteq_cases[consumes 1,elim] = supteqp_cases[to_set]
lemma supteq_supt_eq_conv: "supteq = supt \<union> Id"  by auto

declare conj_cong[fundef_cong]
fun poss :: "('f,'v)term \<Rightarrow> pos set"
where "poss(Var x) = {\<epsilon>}"
    | "poss(Fun f ss) = {\<epsilon>} \<union> {i<#p | i p. i < length ss \<and> p \<in> poss(ss!i)}"
declare conj_cong[fundef_cong del]

fun poss_list :: "('f,'v)term \<Rightarrow> pos list"
where "poss_list (Var x) = [\<epsilon>]"
    | "poss_list(Fun f ss) = (\<epsilon> # concat (map (\<lambda> (i,ps). map ( op <# i) ps) (zip [0 ..< length ss] (map poss_list ss))))"


lemma poss_list_sound[simp] : "set (poss_list s) = poss s" 
proof (induct s, simp)
  case (Fun f ss)
  let ?z = "zip [0..<length ss] (map poss_list ss)"
  have len: "length ?z = length ss" by auto
  have "(\<Union>a\<in>set ?z.
         set (split (\<lambda>i. map (op <# i)) a)) = 
       {i <# p |i p. i < length ss \<and> p \<in> poss (ss ! i)}" (is "?l = ?r")
  proof (rule set_ext)
    fix ip
    show "(ip \<in> ?l) = (ip \<in> ?r)"
    proof
      assume "ip \<in> ?l" 
      from this obtain ipI where 
	z: "ipI \<in> set ?z" and 
	ip: "ip \<in> set (split (\<lambda> i. map (op <# i)) ipI)" 
	by auto     
      from z obtain i where i: "i < length ?z" and zi: "?z ! i = ipI" by (force simp: set_conv_nth)      
      with ip Fun show "ip \<in> ?r" by auto
    next
      assume "ip \<in> ?r"
      from this obtain i p where i: "i < length ss" and "p \<in> poss (ss ! i)"
	and ip: "ip = i <# p" by auto
      with Fun have p: "p \<in> set (poss_list (ss ! i))" and iz: "i < length ?z" by auto
      from i have id: "?z ! i = (i, poss_list (ss ! i))" (is "_ = ?ipI") by auto
      from iz have  "?z ! i \<in> set ?z" by (rule nth_mem)
      with id have inZ: "?ipI \<in> set ?z" by auto
      from p have "i <# p \<in> set (split (\<lambda> i. map (op <# i)) ?ipI)" by auto
      with inZ ip show "ip \<in> ?l" by force
    qed
  qed
  with Fun show ?case by simp
qed

declare poss_list.simps[simp del]

text {* The subterm relation is strongly normalizing. *}
lemma SN_suptp: "SN{(x,y). x \<rhd> y}" unfolding SN_iff_wf
  by (rule wf_subset) (auto intro: suptp_size)
lemmas SN_supt = SN_suptp[to_set]

text {* The subterm of a term~@{text s} at position~@{text p} is defined as follows: *}
fun subt_at :: "('f,'v)term \<Rightarrow> pos \<Rightarrow> ('f,'v)term" ("(_ |'_ _)" [60])
where "s|_\<epsilon> = s"
    | "Fun f ss|_i<#p = ss!i|_p"

text {* A \emph{context} is a term containing exactly one \emph{hole}. *}
datatype ('f,'v)ctxt = Hole ("\<box>") | More 'f "('f,'v)term list" "('f,'v)ctxt" "('f,'v)term list"

text {* Replacing the hole in a context~@{text C} by a term ~@{text t} is called, application
of @{text C} to @{text t}. *}
fun ctxt_apply :: "('f,'v)ctxt \<Rightarrow> ('f,'v)term \<Rightarrow> ('f,'v)term" ("_\<langle>_\<rangle>" [1000, 0] 1000)
where "\<box>\<langle>s\<rangle> = s"
    | "(More f ss1 C ss2)\<langle>s\<rangle> = Fun f (ss1@C\<langle>s\<rangle>#ss2)"

lemma ctxt_eq[simp] : "(C\<langle>s\<rangle> = C\<langle>t\<rangle>) = (s = t)"
  by (induct C, auto)

fun ctxt_of_pos_term :: "pos \<Rightarrow> ('f,'v)term \<Rightarrow> ('f,'v)ctxt"
where "ctxt_of_pos_term \<epsilon> t = \<box>"
    | "ctxt_of_pos_term (i<#ps) (Fun f ts) = More f (take i ts) (ctxt_of_pos_term ps (ts!i)) (drop (Suc i) ts)"

lemma ctxt_supt_id : assumes "p \<in> poss t" 
  shows "(ctxt_of_pos_term p t)\<langle>t|_p\<rangle> = t"
using assms
  by (induct t arbitrary: p, auto simp: id_take_nth_drop[symmetric])

text {* Let @{text s} and @{text t} be terms. The following three statements are equivalent:
\begin{enumerate}
 \item \label{A}@{text "s \<unrhd> t"}
 \item \label{B}@{text "\<exists>p\<in>poss s. s|_p = t"}
 \item \label{C}@{text "\<exists>C. s = C\<langle>t\<rangle>"}
\end{enumerate} *}

fun ctxt_compose :: "('f,'v)ctxt \<Rightarrow> ('f,'v)ctxt \<Rightarrow> ('f,'v)ctxt" (infixl "\<circ>c" 65)
where "\<box> \<circ>c D = D"
    | "(More f ss1 C ss2) \<circ>c D = More f ss1 (C \<circ>c D) ss2"

lemma ctxt_compose_apply_eq[iff]: "(C \<circ>c D)\<langle>t\<rangle> = C\<langle>D\<langle>t\<rangle>\<rangle>" by (induct C) auto

interpretation ctxt_semigroup_mult: semigroup_mult "op \<circ>c :: ('f,'v)ctxt \<Rightarrow> ('f,'v)ctxt \<Rightarrow> ('f,'v)ctxt" proof
 fix C D E :: "('f,'v)ctxt" show "C \<circ>c D \<circ>c E = C \<circ>c (D \<circ>c E)" by (induct C) auto
qed

lemmas ctxt_compose_assoc[simp] = ctxt_semigroup_mult.mult_assoc[symmetric]

interpretation ctxt_monoid_mult: monoid_mult "\<box>" "op \<circ>c" proof
 fix C show "\<box> \<circ>c C = C" by simp
next
 fix C show "C \<circ>c \<box> = C" by (induct C) auto
qed

instantiation ctxt :: (type,type)power begin
  definition one_ctxt_def: "1 = \<box>"
  definition times_ctxt_def: "C * D = C \<circ>c D"
  instance by (intro_classes)
end

text {* The position of the hole in a context is uniquely determined. *}
fun hole_pos :: "('f,'v)ctxt \<Rightarrow> pos"
where "hole_pos \<box> = \<epsilon>"
    | "hole_pos(More _ ss D _) = length ss <# hole_pos D"

lemma supteqp_imp_subt_at: assumes "s \<unrhd> t" shows "\<exists>p\<in>poss s. s|_p = t"
using assms proof (induct s t rule: supteqp.induct)
 case (refl s)
 have "\<epsilon> \<in> poss s" by (induct s rule: term_induct) auto
 have "s|_\<epsilon> = s" by simp
 from `\<epsilon> \<in> poss s` and `s|_\<epsilon> = s` show ?case by best
next
 case (subteq_of_arg u ss t f) 
 then obtain p where "p \<in> poss u" and "u|_p = t" by best
 from `u \<in> set ss` obtain i where "i < length ss" and "u = ss!i" by (auto simp: in_set_conv_nth)
 from `i < length ss` and `p \<in> poss u` have "i<#p \<in> poss (Fun f ss)" unfolding `u = ss!i` by simp
 have "Fun f ss|_i<#p = t" unfolding subt_at.simps unfolding `u = ss!i`[symmetric] by (rule `u|_p = t`)
 with `i<#p \<in> poss (Fun f ss)` show ?case by best
qed

lemma subt_at_imp_ctxt: assumes "p \<in> poss s" shows "\<exists>C. C\<langle>s|_p\<rangle> = s"
using assms proof (induct p arbitrary: s)
 case (Empty s)
 have "\<box>\<langle>s|_\<epsilon>\<rangle> = s" by simp
 thus ?case by best
next
 case (PCons i p s)
 then obtain f ss where "s = Fun f ss" by (cases s) auto
 with `i<#p \<in> poss s` obtain u :: "('a,'b) term"
  where "u = ss!i" and "p \<in> poss u" and "i < length ss" by auto
 from PCons and `p\<in>poss u` obtain D where "D\<langle>u|_p\<rangle> = u" by auto
 let ?ss1 = "take i ss" and ?ss2 = "drop (Suc i) ss"
 let ?E = "More f ?ss1 D ?ss2"
 have "?ss1@D\<langle>u|_p\<rangle>#?ss2 = ss" (is "?ss = ss") unfolding `D\<langle>u|_p\<rangle> = u` unfolding `u = ss!i`
  unfolding id_take_nth_drop[OF `i < length ss`, symmetric] ..
 have "s|_i<#p = u|_p" unfolding `s = Fun f ss` using `u = ss!i` by simp
 have "?E\<langle>s|_i<#p\<rangle> = s" unfolding ctxt_apply.simps `s|_i<#p = u|_p` `?ss = ss` unfolding `s = Fun f ss` ..
 thus ?case by best
qed

lemma ctxt_imp_supteqp: shows "C\<langle>t\<rangle> \<unrhd> t" by (induct C) auto

lemma subt_at_imp_supteqp': assumes "p \<in> poss s" and "s|_p = t" shows "s \<unrhd> t"
proof -
 from `p \<in> poss s` obtain C where "C\<langle>s|_p\<rangle> = s" using subt_at_imp_ctxt by best
 thus ?thesis unfolding `s|_p = t` using ctxt_imp_supteqp by auto
qed

lemma subt_at_imp_supteqp: assumes "p \<in> poss s" shows "s \<unrhd> s|_p"
proof -
 have "s|_p = s|_p" by auto
 with assms show ?thesis by (rule subt_at_imp_supteqp')
qed


(* Conversions between contexts and subterms. *)
lemma supteqp_ctxt_E: assumes "s \<unrhd> t" obtains C where "s = C\<langle>t\<rangle>"
using assms proof (induct arbitrary: thesis)
  case (refl s)
  have "s = \<box>\<langle>s\<rangle>" by simp
  from refl[OF this] show ?case .
next
  case (subteq_of_arg u ss t f)
  then obtain C where "u = C\<langle>t\<rangle>" by auto
  from split_list[OF `u \<in> set ss`] obtain ss1 and ss2 where "ss = ss1 @ u # ss2" by auto
  hence "Fun f ss = (More f ss1 C ss2)\<langle>t\<rangle>" using `u = C\<langle>t\<rangle>` by simp
  with subteq_of_arg show ?case by best
qed

lemma ctxt_supteqp_I: assumes "s = C\<langle>t\<rangle>" shows "s \<unrhd> t"
proof (cases C)
  case Hole thus ?thesis using assms by auto
next
  case (More f ss1 D ss2)
  with assms have s: "s = Fun f (ss1 @ D\<langle>t\<rangle> # ss2)" (is "_ = Fun _ ?ss") by simp
  have "D\<langle>t\<rangle> \<in> set ?ss" by simp
  moreover have "D\<langle>t\<rangle> \<unrhd> t" by (induct D) auto
  ultimately show ?thesis unfolding s ..
qed

lemma supteqp_ctxt_conv: "(s \<unrhd> t) = (\<exists>C. s = C\<langle>t\<rangle>)" (is "?lhs = ?rhs")
proof
  assume "?lhs" thus "?rhs" by (auto elim: supteqp_ctxt_E)
next
  assume "?rhs" thus "?lhs" by (auto intro: ctxt_supteqp_I)
qed

(* Conversions between contexts and proper subterms. *)
lemma suptp_ctxt_E: assumes "s \<rhd> t" obtains C where "C \<noteq> \<box>" and "s = C\<langle>t\<rangle>"
using assms proof (induct arbitrary: thesis)
  case (arg s ss f)
  from split_list[OF `s \<in> set ss`] obtain ss1 and ss2 where ss: "ss = ss1 @ s # ss2" by auto
  let ?C = "More f ss1 \<box> ss2"
  have "?C \<noteq> \<box>" by simp
  moreover have "Fun f ss = ?C\<langle>s\<rangle>" by (simp add: ss)
  ultimately show ?case using arg by best
next
  case (subt_of_arg s ss t f)
  then obtain C where "C \<noteq> \<box>" and "s = C\<langle>t\<rangle>" by auto
  from split_list[OF `s \<in> set ss`] obtain ss1 and ss2 where ss: "ss = ss1 @ s # ss2" by auto
  have "More f ss1 C ss2 \<noteq> \<box>" by simp
  moreover have "Fun f ss = (More f ss1 C ss2)\<langle>t\<rangle>" using `s = C\<langle>t\<rangle>` by (simp add: ss)
  ultimately show ?case using subt_of_arg(4) by best
qed

(* More transitivity rules. *)
lemma suptp_supteqp[trans]:
  assumes "s \<rhd> t" and "t \<unrhd> u" shows "s \<rhd> u"
proof (cases "t = u")
  assume "t = u" with assms show ?thesis by simp
next
  assume "t \<noteq> u"
  with assms have "t \<rhd> u" by auto
  with `s \<rhd> t` show ?thesis by (rule suptp_trans)
qed

lemma supteqp_suptp[trans]:
  assumes "s \<unrhd> t" and "t \<rhd> u" shows "s \<rhd> u"
proof (cases "s = t")
  assume "s = t" with assms show ?thesis by simp
next
  assume "s \<noteq> t"
  with assms have "s \<rhd> t" by auto
  thus ?thesis using `t \<rhd> u` by (rule suptp_trans)
qed

lemma ctxt_suptp_I: assumes "C \<noteq> \<box>" and "s = C\<langle>t\<rangle>" shows "s \<rhd> t"
proof (cases C)
  case Hole with assms show ?thesis by simp
next
  case (More f ss1 D ss2)
  with assms have s: "s = Fun f (ss1 @ D\<langle>t\<rangle> # ss2)" by simp
  have "D\<langle>t\<rangle> \<in> set (ss1 @ D\<langle>t\<rangle> # ss2)" by simp
  hence "s \<rhd> D\<langle>t\<rangle>" unfolding s ..
  also have "\<dots> \<unrhd> t" by (induct D) auto
  finally show "s \<rhd> t" .
qed

lemma suptp_ctxt_conv: "(s \<rhd> t) = (\<exists>C. C \<noteq> \<box> \<and> s = C\<langle>t\<rangle>)" (is "?lhs = ?rhs")
proof
  assume "?lhs"
  from suptp_ctxt_E[OF this] show "?rhs" by blast
next
  assume "?rhs"
  then obtain C where "C \<noteq> \<box>" and "s = C\<langle>t\<rangle>" by auto
  thus "?lhs" by (rule ctxt_suptp_I)
qed

lemma supteqp_imp_ctxt: assumes "s \<unrhd> t" shows "\<exists>C. C\<langle>t\<rangle> = s"
proof -
 from assms obtain p where "p \<in> poss s" and "s|_p = t" using supteqp_imp_subt_at by best
 then obtain C where "C\<langle>t\<rangle> = s" unfolding `s|_p = t`[symmetric] using subt_at_imp_ctxt by best
 thus ?thesis by best
qed

text {* By adding \emph{non-empty} to statements \ref{B} and \ref{C} a similar characterisation for
proper subterms is obtained:
\begin{enumerate}
 \item @{text "s \<rhd> t"}
 \item @{text "\<exists>i p. i#p\<in>poss s \<and> s|_i<#p = t"}
 \item @{text "\<exists>C. C \<noteq> \<box> \<and> s = C\<langle>t\<rangle>"}
\end{enumerate} *}

lemma suptp_imp_subt_at_nepos: assumes "s \<rhd> t" shows "\<exists>i p. i<#p\<in>poss s \<and> s|_i<#p = t"
proof -
 from assms have "s \<unrhd> t" and "s \<noteq> t" unfolding supt_iff_supteq_neq by auto
 then obtain p where supteq: "p \<in> poss s" "s|_p = t" using supteqp_imp_subt_at by best
 have "p \<noteq> \<epsilon>"
 proof
  assume "p = \<epsilon>" hence "s = t" using `s|_p = t` by simp
  thus False using `s \<noteq> t` by simp
 qed
 then obtain i q where "p = i<#q" by (cases p) simp
 with supteq show ?thesis by auto
qed

lemma subt_at_nepos_neq: assumes "i<#p \<in> poss s" shows "s|_(i<#p) \<noteq> s"
proof (cases s)
 fix x assume "s = Var x"
 hence "i<#p \<notin> poss s" by simp
 with assms show ?thesis by simp
next
 fix f ss assume "s = Fun f ss" show ?thesis
 proof
  assume "s|_i<#p = s"
  from assms have "i < length ss" unfolding `s = Fun f ss` by auto
  hence "ss!i \<in> set ss" by simp
  hence "Fun f ss \<rhd> ss!i" by (rule suptp.arg)
  hence "Fun f ss \<noteq> ss!i" unfolding supt_iff_supteq_neq by simp
  from `s|_i<#p = s` and assms
  have "ss!i \<unrhd> Fun f ss" using subt_at_imp_supteqp' unfolding `s = Fun f ss` by auto
  thus False
  proof (cases rule: supteqp_cases)
   assume "ss!i = Fun f ss" with `Fun f  ss \<noteq> ss!i` show ?thesis by simp
  next
   assume "ss!i \<rhd> Fun f ss" with `Fun f ss \<rhd> ss!i` show ?thesis using suptp_asym by auto
  qed
 qed
qed


lemma subt_at_nepos_imp_suptp: assumes "i<#p \<in> poss s" shows "s \<rhd> s|_i<#p"
proof -
 from assms have "s \<unrhd> s|_i<#p" by (rule subt_at_imp_supteqp)
 from assms have "s|_i<#p \<noteq> s" by (rule subt_at_nepos_neq)
 from `s \<unrhd> s|_i<#p` and `s|_i<#p \<noteq> s` show ?thesis by (auto simp: supt_iff_supteq_neq)
qed

lemma subt_at_nepos_imp_nectxt: assumes "i<#p \<in> poss s" and "s|_i<#p = t" shows "\<exists>C. C \<noteq> \<box> \<and> C\<langle>t\<rangle> = s"
proof -
 from assms obtain C where "C\<langle>s|_i<#p\<rangle> = s" using subt_at_imp_ctxt by best
 from `i<#p \<in> poss s` have "t \<noteq> s" unfolding `s|_i<#p = t`[symmetric] using subt_at_nepos_neq by best
 from assms and `C\<langle>s|_i<#p\<rangle> = s` have "C\<langle>t\<rangle> = s" by simp
 have "C \<noteq> \<box>"
 proof
  assume "C = \<box>"
  with `C\<langle>t\<rangle> = s` have "t = s" by simp
  with `t \<noteq> s` show False by simp
 qed
 with `C\<langle>t\<rangle> = s` show ?thesis by best
qed

lemma nectxt_imp_suptp_ctxt: "C \<noteq> \<box> \<Longrightarrow> C\<langle>t\<rangle> \<rhd> t"
proof (induct C)
 case Hole thus ?case by auto
next
 case (More f ls C rs) thus ?case
 proof (cases "C = \<box>")
  case True
  hence a: "(More f ls C rs)\<langle>t\<rangle> = Fun f (ls @ t#rs)" by auto
  hence "t \<in> set (ls @ t#rs)" by auto
  hence "Fun f (ls@t#rs) \<rhd> t" by (rule suptp.arg)
  with a show ?thesis by auto
 next
  case False
  from `C \<noteq> \<box>` and `C \<noteq> \<box> \<Longrightarrow> C\<langle>t\<rangle> \<rhd> t` have b: "C\<langle>t\<rangle> \<rhd> t" by auto
  have c: "(More f ls C rs)\<langle>t\<rangle> = Fun f (ls @ C\<langle>t\<rangle>#rs)" by auto
  have d: "C\<langle>t\<rangle> \<in> set (ls @ C\<langle>t\<rangle>#rs)" by auto
  have "Fun f (ls @ C\<langle>t\<rangle>#rs) \<rhd> t" using suptp.arg[where s = "C\<langle>t\<rangle>" and ss = "ls@C\<langle>t\<rangle>#rs",OF d] b by (rule suptp_trans)
  with c show ?thesis by auto
 qed
qed

lemma nectxt_imp_suptp: 
 assumes A:"C \<noteq> \<box>" and B: "t = C\<langle>s\<rangle>"
 shows "t \<rhd> s"
proof -
 from A have "C\<langle>s\<rangle> \<rhd> s" by (rule nectxt_imp_suptp_ctxt)
 with B show ?thesis by auto
qed

lemma suptp_imp_nectxt: assumes "s \<rhd> t" shows "\<exists>C. C \<noteq> \<box> \<and> C\<langle>t\<rangle> = s"
proof -
 from assms have "s \<unrhd> t" and "s \<noteq> t" using supt_iff_supteq_neq by auto
 then obtain C where "C\<langle>t\<rangle> = s" using supteqp_imp_ctxt by best
 have "C \<noteq> \<box>"
 proof
  assume "C = \<box>" hence "C\<langle>t\<rangle> = t" by simp
  hence "s = t" unfolding `C\<langle>t\<rangle> = s` .
  with `s \<noteq> t` show False by simp
 qed
 with `C\<langle>t\<rangle> = s` show ?thesis by auto
qed

lemma suptp_var: "\<not> (Var x \<rhd> u)"
proof
  assume "Var x \<rhd> u"
  with suptp_imp_nectxt obtain C where "C \<noteq> \<box> \<and> C\<langle>u\<rangle> = Var x" by auto
  thus False by (cases C, auto)
qed

lemma suptp_const: "\<not> (Fun f [] \<rhd> u)"
proof
  assume "Fun f [] \<rhd> u"
  with suptp_imp_nectxt obtain C where "C \<noteq> \<box> \<and> C\<langle>u\<rangle> = Fun f []" by (auto simp: Nil)
  thus False by (cases C, auto)
qed

lemma supteqp_var_imp_eq :
  "(Var x \<unrhd> t) = (t = Var x)" (is "_ = (_ = ?x)")
proof
  assume "t = Var x" thus "Var x \<unrhd> t" by auto
next
  assume st: "?x \<unrhd> t"
  from st supteqp_imp_ctxt obtain C where  "C\<langle>t\<rangle> = ?x" by force
  thus "t = ?x" by (cases C) auto
qed

lemma subterm_induct: 
  assumes base: "\<And>x. P (Var x)"
    and step: "\<And>f ss. (\<And>s u. s \<in> set ss \<and> s \<unrhd> u \<Longrightarrow> P u) \<Longrightarrow> P (Fun f ss)" 
  shows [case_names Var Fun]: "P t"
proof -
  have "\<forall>s\<unlhd>t. P s"
  proof (induct t)
    case (Var x) thus ?case by (simp add: base supteqp_var_imp_eq)
  next
    case (Fun f ss)
    from Fun step have P: "P (Fun f ss)" by blast
    show ?case
    proof (intro impI allI)
      fix s
      assume "Fun f ss \<unrhd> s"
      thus "P s"
      proof
        assume "Fun f ss = s" with P show "P s" by simp
      next
        assume s: "Fun f ss \<rhd> s"
        obtain C where "Fun f ss = C\<langle>s\<rangle>" and "C \<noteq> \<box>" using suptp_imp_nectxt[OF s] by force
        then obtain b a D where "Fun f ss = Fun f (b @ (D\<langle>s\<rangle>) # a)" by (cases C, auto)
        hence mem: "D\<langle>s\<rangle> \<in> set ss" by auto
        from Fun[OF mem] ctxt_imp_supteqp[of D s] show "P s" by simp
      qed
    qed
  qed
  thus ?thesis by auto
qed


text {* A \emph{substitution} is a mapping~$\sigma$ from variables to terms. *}
(* A new datatype is created in order to support instanciation of type classes. *)
datatype ('f,'v)sub = Sub "'v \<Rightarrow> ('f,'v)term"

fun get_sub where "get_sub (Sub \<sigma>) = \<sigma>"

abbreviation sub_empty :: "('f,'v)sub" ("\<iota>") where "\<iota> \<equiv> Sub(\<lambda>x. Var x)"
fun dom :: "('f,'v)sub \<Rightarrow> 'v set" where "dom(Sub \<sigma>) = {x. \<sigma>(x) \<noteq> (Var x)}"

fun ran :: "('f,'v)sub \<Rightarrow> ('f,'v)term set" where "ran(Sub \<sigma>) = \<sigma>`(dom(Sub \<sigma>))"

fun sub_restrict :: "('f,'v)sub \<Rightarrow> 'v set \<Rightarrow> ('f,'v)sub" ("_|'__" 60)
where "(Sub \<sigma>)|_V = Sub(\<lambda>x.(if x \<in> V then \<sigma> x else (Var x)))"

lemma sub_empty_empty_dom[simp]: "dom \<iota> = {}" by simp

lemma sub_empty_conv[iff]: fixes \<sigma> :: "('f,'v)sub" shows "(dom \<sigma> = {}) = (\<sigma> = \<iota>)"
proof (rule iffI,cases \<sigma>)
 case (Sub s)
 assume "dom \<sigma> = {}"
 have "s = (\<lambda>x. Var x)"
 proof (rule ext)
  fix x show "s x = Var x"
  proof (rule ccontr)
   assume "s x \<noteq> Var x"
   hence "x \<in> dom \<sigma>" unfolding Sub by simp
   with `dom \<sigma> = {}` show False by simp
  qed
 qed
 thus "\<sigma> = \<iota>" unfolding Sub by simp
next
 assume "\<sigma> = \<iota>" thus "dom \<sigma> = {}" by simp
qed

fun sub_apply :: "('f,'v)term \<Rightarrow> ('f,'v)sub \<Rightarrow> ('f,'v)term" (infixl "\<cdot>" 60)
where "(Var x)\<cdot>(Sub \<sigma>) = \<sigma> x"
    | "(Fun f ss)\<cdot>\<sigma> = Fun f (map (\<lambda>t. t\<cdot>\<sigma>) ss)"

fun sub_compose :: "('f,'v)sub \<Rightarrow> ('f,'v)sub \<Rightarrow> ('f,'v)sub" (infixl "\<circ>s" 65)
where "(Sub \<sigma>) \<circ>s \<tau> = Sub(\<lambda>x. (\<sigma> x)\<cdot>\<tau>)"

lemma term_sub_distr:
 fixes t :: "('f,'v)term" and \<sigma> \<tau> :: "('f,'v)sub" shows "(t\<cdot>\<sigma>)\<cdot>\<tau> = t\<cdot>(\<sigma> \<circ>s \<tau>)"
proof (induct t)
  case (Var x) thus ?case by (cases \<sigma>) auto
next
  case (Fun f ss)
  hence IH:"\<And>s. s \<in> set ss \<longrightarrow> (s\<cdot>\<sigma>)\<cdot>\<tau> = s\<cdot>(\<sigma> \<circ>s \<tau>)" by auto
  have IH': "map (\<lambda>s. (s\<cdot>\<sigma>)\<cdot>\<tau>) ss = map (\<lambda>s. s\<cdot>(\<sigma> \<circ>s \<tau>)) ss" using map_ext[OF IH] .
  have "((Fun f ss)\<cdot>\<sigma>)\<cdot>\<tau> = (Fun f (map (\<lambda>t. t\<cdot>\<sigma>) ss))\<cdot>\<tau>" by simp
  also have "\<dots> = Fun f (map (\<lambda>t. t\<cdot>\<tau>) (map (\<lambda>t. t\<cdot>\<sigma>) ss))" by simp
  also have "\<dots> = Fun f (map ((\<lambda>t. t\<cdot>\<tau>)\<circ>(\<lambda>t. t\<cdot>\<sigma>)) ss)" by simp
  also have "\<dots> = Fun f (map (\<lambda>t. (t\<cdot>\<sigma>)\<cdot>\<tau>) ss)" by simp
  also have "\<dots> = Fun f (map (\<lambda>t. t\<cdot>(\<sigma> \<circ>s \<tau>)) ss)" unfolding IH' by simp
  finally show ?case by simp
qed

lemmas term_sub_compose_distr[simp] = term_sub_distr[symmetric]

lemma sub_compose_assoc[simp]: fixes \<sigma> \<tau> \<upsilon> :: "('f,'v)sub" shows "\<sigma> \<circ>s \<tau> \<circ>s \<upsilon> = \<sigma> \<circ>s (\<tau> \<circ>s \<upsilon>)"
proof (cases \<sigma>)
 case (Sub s)
 have "\<sigma> \<circ>s \<tau> \<circ>s \<upsilon> = (Sub s) \<circ>s \<tau> \<circ>s \<upsilon>" unfolding Sub by simp
 also have "\<dots> = (Sub(\<lambda>x. (s x)\<cdot>\<tau>)) \<circ>s \<upsilon>" by simp
 also have "\<dots> = Sub(\<lambda>x. ((s x)\<cdot>\<tau>)\<cdot>\<upsilon>)" by simp
 also have "\<dots> = Sub(\<lambda>x. (s x)\<cdot>(\<tau> \<circ>s \<upsilon>))" by simp
 finally show ?thesis unfolding Sub by simp
qed

interpretation sub_semigroup: semigroup_mult "op \<circ>s :: ('f,'v)sub \<Rightarrow> ('f,'v)sub \<Rightarrow> ('f,'v)sub" proof
 fix \<sigma> \<tau> \<upsilon> show "\<sigma> \<circ>s \<tau> \<circ>s \<upsilon> = \<sigma> \<circ>s (\<tau> \<circ>s \<upsilon>)" by (rule sub_compose_assoc)
qed

lemma sub_apply_id[simp]: "t\<cdot>\<iota> = t"
proof (induct t)
 case (Var x) show ?case by simp
next
 case (Fun f ss)
 have IH: "map (\<lambda>t. t\<cdot>\<iota>) ss = ss" by (rule map_idI[OF Fun])
 show ?case unfolding sub_apply.simps unfolding IH by simp
qed

interpretation sub: monoid_mult "\<iota> :: ('f,'v)sub" "op \<circ>s :: ('f,'v)sub \<Rightarrow> ('f,'v)sub \<Rightarrow> ('f,'v)sub" proof
 fix \<sigma> show "\<iota> \<circ>s \<sigma> = \<sigma>" by (cases \<sigma>) simp
next
 fix \<sigma> show "\<sigma> \<circ>s \<iota> = \<sigma>" by (cases \<sigma>) simp
qed

instantiation sub :: (type,type)power begin
  definition one_sub_def: "1 = \<iota>"
  definition times_sub_def: "\<sigma> * \<tau> = \<sigma> \<circ>s \<tau>"
  instance by (intro_classes)
end

lemma [simp]: fixes ts :: "('f,'v)term list" shows "map (\<lambda>t. t\<cdot>\<iota>) ts = ts"
proof (induct ts)
 case Nil thus ?case by simp
next
 case (Cons t ts) 
 have "map (\<lambda>t. t\<cdot>\<iota>) (t#ts) = (t\<cdot>\<iota>) # map (\<lambda>t. t\<cdot>\<iota>) ts" by simp
 also have "\<dots> = t # map (\<lambda>t. t\<cdot>\<iota>) ts" by simp
 finally show ?case using Cons by simp
qed

fun sub_apply_to_ctxt :: "('f,'v)ctxt \<Rightarrow> ('f,'v)sub \<Rightarrow> ('f,'v)ctxt" (infixl "\<cdot>c" 60)
where "\<box> \<cdot>c \<sigma> = \<box>"
    | "(More f ss1 D ss2) \<cdot>c \<sigma> = More f (map (\<lambda>t. t\<cdot>\<sigma>) ss1) (D \<cdot>c \<sigma>) (map (\<lambda>t. t\<cdot>\<sigma>) ss2)"

lemma ctxt_sub_distr[simp]: "C\<langle>t\<rangle>\<cdot>\<mu> = (C \<cdot>c \<mu>)\<langle>t\<cdot>\<mu>\<rangle>" by (induct C) auto

lemma ctxt_compose_sub_distr[simp]: "(C \<circ>c D) \<cdot>c \<sigma> = (C \<cdot>c \<sigma>) \<circ>c (D \<cdot>c \<sigma>)" by (induct C) auto

lemma [simp]: "map (\<lambda>t. (t \<cdot> \<sigma>) \<cdot> \<tau>) ss = map (\<lambda>t. t \<cdot> \<tau>) (map (\<lambda>t. t \<cdot> \<sigma>) ss)" by (induct ss) auto

lemma sub_compose_ctxt_distr[simp]: "C \<cdot>c (\<sigma> \<circ>s \<tau>) = (C \<cdot>c \<sigma>) \<cdot>c \<tau>" by (induct C) auto

text {* Context substitutions. *}
datatype ('f,'v)cs = CS "('f,'v)ctxt" "('f,'v)sub"

fun cs_apply :: "('f,'v)term \<Rightarrow> ('f,'v)cs \<Rightarrow> ('f,'v)term" (infixl "\<cdot>cs" 60)
where "t \<cdot>cs (CS C \<mu>) = C\<langle>t\<cdot>\<mu>\<rangle>"

fun cs_compose :: "('f,'v)cs \<Rightarrow> ('f,'v)cs \<Rightarrow> ('f,'v)cs" (infixl "\<circ>cs" 60)
where "(CS C \<sigma>) \<circ>cs (CS D \<tau>) = CS (D \<circ>c (C \<cdot>c \<tau>)) (\<sigma> \<circ>s \<tau>)"

lemma id_sub_ctxt[simp]: fixes C :: "('f,'v)ctxt" shows "C \<cdot>c \<iota> = C" by (induct C) auto

lemma cs_compose_id_right[simp]: "(CS C \<sigma>) \<circ>cs (CS \<box> \<iota>) = (CS C \<sigma>)"
proof -
 have "(CS C \<sigma>) \<circ>cs (CS \<box> \<iota>) = CS (\<box> \<circ>c (C \<cdot>c \<iota>)) (\<sigma> \<circ>s \<iota>)" by simp
 also have "\<dots> = CS C (\<sigma> \<circ>s \<iota>)" by simp
 also have "\<dots> = CS C \<sigma>" by simp
 finally show ?thesis by simp
qed

lemma cs_compose_left_neutral[simp]: "(CS \<box> \<iota>) \<circ>cs (CS C \<sigma>) = CS C \<sigma>"
proof -
 have "CS \<box> \<iota> \<circ>cs CS C \<sigma> = CS (C \<circ>c (\<box> \<cdot>c \<sigma>)) (\<iota> \<circ>s \<sigma>)" by simp
 also have "\<dots> = CS (C \<circ>c \<box>) (\<iota> \<circ>s \<sigma>)" by simp
 also have "\<dots> = CS C (\<iota> \<circ>s \<sigma>)" by simp
 also have "\<dots> = CS C \<sigma>" by (cases \<sigma>) simp
 finally show ?thesis by simp
qed

interpretation cs: semigroup_mult "op \<circ>cs :: ('f,'v)cs \<Rightarrow> ('f,'v)cs \<Rightarrow> ('f,'v)cs" proof
 fix c d e :: "('f,'v)cs" show "c \<circ>cs d \<circ>cs e = c \<circ>cs (d \<circ>cs e)" by (cases c,cases d,cases e) simp
qed

interpretation cs: monoid_mult "(CS \<box> \<iota>)::('f,'v)cs" "op \<circ>cs" proof
 fix c show "(CS \<box> \<iota>) \<circ>cs c = c" by (cases c) (simp only: cs_compose_left_neutral)
next
 fix c show "c \<circ>cs (CS \<box> \<iota>) = c" by (cases c) simp
qed

instantiation cs :: (type,type)power begin
  definition one_cs_def: "1 = CS \<box> \<iota>"
  definition times_cs_def: "Cs * Ds = Cs \<circ>cs Ds"
  instance by (intro_classes)
end

lemma term_cs_0_power[simp]: "t \<cdot>cs c^(0::nat) = t" by (simp add: one_cs_def)

lemma term_cs_compose_distr[simp]: "t \<cdot>cs (c \<circ>cs d) = (t \<cdot>cs c) \<cdot>cs d" by (cases c,cases d,induct t) auto

lemma sub_power_compose_distr: fixes \<mu> :: "('f,'v)sub" shows "\<mu>^(m+n) = (\<mu>^(m::nat) \<circ>s \<mu>^n)"
proof (cases \<mu>)
 case (Sub s) show ?thesis unfolding Sub
 proof (induct m)
  case 0
  have "Sub s ^(0+n) = Sub s ^ n" by simp
  also have "\<dots> = \<iota> \<circ>s (Sub s)^n" unfolding sub.mult_1_left by simp
  finally show ?case by (simp add: one_sub_def)
 next
  case (Suc n) thus ?case by (simp add: times_sub_def)
 qed
qed

lemma cs_power_compose_distr: "(CS C \<mu>)^(m+n) = (CS C \<mu>)^(m::nat) \<circ>cs (CS C \<mu>)^n"
proof (induct m arbitrary: n)
 case 0 thus ?case by (simp add: one_cs_def)
next
 case (Suc m)
 have "(CS C \<mu>)^(Suc m + n) = (CS C \<mu>)^Suc(m+n)" by simp
 also have "\<dots> = (CS C \<mu>) \<circ>cs (CS C \<mu>)^(m+n)" by (simp add: times_cs_def)
 also have "\<dots> = (CS C \<mu>) \<circ>cs ((CS C \<mu>)^m \<circ>cs (CS C \<mu>)^n)" unfolding Suc by simp
 also have "\<dots> = (CS C \<mu>) \<circ>cs (CS C \<mu>)^m \<circ>cs (CS C \<mu>)^n" by (simp add: cs.mult_assoc)
 finally show ?case by (simp add: times_cs_def)
qed

lemma term_cs_induct_power[simp]: "t \<cdot>cs (CS C \<mu>)^Suc(n::nat) = C\<langle>t \<cdot>cs (CS C \<mu>)^n \<cdot> \<mu>\<rangle>"
proof -
 have "t \<cdot>cs (CS C \<mu>)^(Suc n) = t \<cdot>cs (CS C \<mu>)^(n + 1)" by simp
 also have "\<dots> = t \<cdot>cs ((CS C \<mu>)^n \<circ>cs (CS C \<mu>))" unfolding cs_power_compose_distr by (simp add: one_cs_def times_cs_def)
 finally show ?thesis by simp
qed

text {* Properties of context-substitutions. *}
lemma cs_sub_distr[simp]: "t \<cdot>cs (CS C \<mu>) ^ n \<cdot> \<mu> = (t\<cdot>\<mu>) \<cdot>cs (CS (C \<cdot>c \<mu>) \<mu>) ^ n"
proof (induct n)
 case 0 thus ?case by (simp add: one_cs_def)
next
 case (Suc n)
 have "t \<cdot>cs (CS C \<mu>)^Suc n \<cdot> \<mu> = C\<langle>t \<cdot>cs (CS C \<mu>)^n \<cdot> \<mu>\<rangle> \<cdot> \<mu>" unfolding term_cs_induct_power by simp
 also have "\<dots> = C\<langle>(t\<cdot>\<mu>) \<cdot>cs (CS (C \<cdot>c \<mu>) \<mu>)^n\<rangle>\<cdot>\<mu>" unfolding Suc by simp
 also have "\<dots> = (C \<cdot>c \<mu>)\<langle>(t\<cdot>\<mu>) \<cdot>cs (CS (C \<cdot>c \<mu>) \<mu>)^n \<cdot> \<mu>\<rangle>" by simp
 finally show ?case unfolding term_cs_induct_power by simp
qed

lemma subt_at_PCons_distr[simp]:
 assumes "i<#p \<in> poss t" shows "t|_i<#p = (t|_<i>)|_p" using assms by (induct t) auto

lemma subt_at_append_distr[simp]: assumes "p<#>q \<in> poss t" shows "t|_(p<#>q) = (t|_p)|_q"
using assms proof (induct p arbitrary: t)
 case Empty thus ?case by simp
next
 case (PCons i p) thus ?case
 proof (induct t)
  case (Var x)
  have "poss(Var x) = {\<epsilon>}" by simp
  from Var have "i<#(p<#>q) = \<epsilon>" unfolding `poss(Var x) = {\<epsilon>}` by simp
  have "i<#(p<#>q) \<noteq> \<epsilon>" by simp
  from this `i<#(p<#>q) = \<epsilon>` show ?case by (rule notE)
 next
  case (Fun f ss) thus ?case by auto
 qed
qed

lemma empty_pos_in_poss[simp]: "\<epsilon> \<in> poss t" by (induct t) auto

lemma hole_pos_poss[simp]: "hole_pos C \<in> poss(C\<langle>t\<rangle>)" by (induct C) auto

lemma hole_pos_ctxt_poss_conv[iff]: "((hole_pos C <#> q) \<in> poss(C\<langle>t\<rangle>)) = (q \<in> poss t)" by (induct C) auto

lemma hole_pos_power_poss[simp]: "(hole_pos C)^(n::nat) \<in> poss((C^n)\<langle>t\<rangle>)" by (induct n) (auto simp: one_ctxt_def one_pos_def times_ctxt_def times_pos_def)

lemma sub_subt_at_distr[simp]: assumes "p \<in> poss t" shows "t|_p \<cdot> \<sigma> = (t \<cdot> \<sigma>)|_p"
using assms by (induct t arbitrary: p) auto

lemma subt_at_hole_pos[simp]: "C\<langle>t\<rangle>|_(hole_pos C) = t" by (induct C) auto

lemma poss_imp_sub_poss[simp]: "p \<in> poss t \<Longrightarrow> p \<in> poss(t\<cdot>\<sigma>)" by (induct t arbitrary: p) auto

lemma poss_imp_ctxt_sub_poss[simp]: assumes "p \<in> poss(C\<langle>t\<rangle>)" shows "p \<in> poss((C \<cdot>c \<sigma>)\<langle>t \<cdot> \<sigma>\<rangle>)"
proof -
 have "p \<in> poss(C\<langle>t\<rangle> \<cdot> \<sigma>)" by (rule poss_imp_sub_poss[OF assms])
 thus ?thesis by simp
qed

lemma cs_hole_ctxt_poss[simp]: "q \<in> poss t \<Longrightarrow> (hole_pos C <#> q) \<in> poss(t \<cdot>cs (CS C \<sigma>))" by simp

lemma cs_hole_poss_power_poss[simp]: "(hole_pos C)^(n::nat) \<in> poss(t \<cdot>cs (CS C \<sigma>)^n)" (is "?p^n \<in> poss(t \<cdot>cs ?c^n)")
proof (induct n)
 case 0 show ?case by (simp add: one_pos_def one_cs_def)
next
 case (Suc n) have "(?p <#> ?p^n) \<in> poss(t \<cdot>cs ?c^n \<cdot>cs ?c)" using cs_hole_ctxt_poss[OF Suc] by best
 hence "?p^(1+n) \<in> poss(t \<cdot>cs (?c^n \<circ>cs ?c^(1::nat)))" unfolding power_append_distr term_cs_compose_distr[symmetric]
  by (simp add: one_cs_def one_pos_def times_pos_def times_cs_def)
 thus ?case unfolding power_append_distr[symmetric] cs_power_compose_distr[symmetric] by simp
qed

lemma subt_at_hole_pos_power[simp]:
 fixes t :: "('f,'v)term"
 shows "((t \<cdot>cs ((CS C \<mu>)^(n::nat)))|_((hole_pos C)^n)) = (t \<cdot> (\<mu>^n))" (is "?A n = ?B n")
proof (induct n)
 case 0 show ?case by (simp add: one_sub_def one_cs_def one_pos_def)
next
 case (Suc n)
 let ?p = "hole_pos C"
 let ?c = "CS C \<mu>"
 have "?p^Suc n \<in> poss(t \<cdot>cs ?c^Suc n)" by (rule cs_hole_poss_power_poss)
 hence "?p^(1 + n) \<in> poss(t \<cdot>cs ?c^(n + 1))" by simp
 hence A: "?p <#> ?p^n \<in> poss(C\<langle>t \<cdot>cs ?c^n \<cdot> \<mu>\<rangle>)" unfolding power_append_distr cs_power_compose_distr
   by (simp add: times_pos_def times_cs_def one_pos_def one_cs_def)
 have "?A(Suc n) = ((t \<cdot>cs ((?c)^(n+1)))|_((?p)^(1+n)))" by simp
 also have "\<dots> = ((t \<cdot>cs ?c^n) \<cdot>cs ?c)|_(?p <#> (?p^n))" unfolding cs_power_compose_distr power_append_distr
   by (simp add: times_cs_def one_cs_def one_pos_def times_pos_def)
 also have "\<dots> = C\<langle>t \<cdot>cs ?c^n \<cdot> \<mu>\<rangle>|_(?p <#> (?p^n))" by simp
 also have "\<dots> = (t \<cdot>cs ?c^n \<cdot> \<mu>)|_(?p^n)" unfolding subt_at_append_distr[OF A] by simp
 also have "\<dots> = (t \<cdot>cs ?c^n)|_(?p^n) \<cdot> \<mu>" by simp
 also have "\<dots> = (t \<cdot> \<mu>^n)\<cdot>\<mu>" unfolding Suc by simp
 also have "\<dots> = t \<cdot> (\<mu>^n \<circ>s \<mu>^1)" unfolding term_sub_distr by (simp add: times_sub_def one_sub_def)
 finally show ?case unfolding sub_power_compose_distr[symmetric] by simp
qed

definition subst :: "'v \<Rightarrow> ('f,'v) term \<Rightarrow> ('f,'v)sub" ("{_\<leftarrow>_}") 
  where "{x\<leftarrow>t} = Sub(\<lambda>y. if x = y then t else Var y)"

inductive sub_instance :: "('f,'v)term \<Rightarrow> ('f,'v) term \<Rightarrow> bool" ("(_/ \<preceq> _)" [51,51] 50)
where sub_instanceI[intro]: "s\<cdot>\<sigma> = t \<Longrightarrow> s \<preceq> t"

lemma sub_instance_trans[trans]:
 fixes s :: "('f,'v)term"
 assumes "s \<preceq> t" and "t \<preceq> u" shows "s \<preceq> u"
proof -
 from `s \<preceq> t` obtain \<sigma> where "s\<cdot>\<sigma> = t" by (cases rule: sub_instance.cases) best
 from `t \<preceq> u` obtain \<tau> where "t\<cdot>\<tau> = u" by (cases rule: sub_instance.cases) best
 hence "(s\<cdot>\<sigma>)\<cdot>\<tau> = u" unfolding `s\<cdot>\<sigma> = t` .
 hence "s\<cdot>(\<sigma> \<circ>s \<tau>) = u" by simp
 thus ?thesis by (rule sub_instanceI)
qed
 
lemma sub_instance_refl: "s \<preceq> s" using sub_instanceI[where \<sigma> = "\<iota>" and s = s and t = s] by simp

lemma sub_neutral: "dom \<sigma> \<subseteq> V \<Longrightarrow> (Var x)\<cdot>(\<sigma>|_(V - {x})) = (Var x)" by (cases \<sigma>) auto

lemma sub_restrict_dom': "(\<lambda>x. if x \<in> dom(Sub s) then s x else (Var x)) x = s x" by simp

lemma sub_restrict_dom[simp]: "\<sigma>|_dom \<sigma> = \<sigma>"
proof (cases \<sigma>)
 case (Sub s)
 have "\<sigma>|_dom \<sigma> = (Sub s)|_dom(Sub s)" unfolding Sub by simp
 also have "\<dots> = Sub(\<lambda>x. if x \<in> dom(Sub s) then s x else (Var x))" by simp
 also have "\<dots> = Sub s" unfolding sub_restrict_dom' by simp
 finally show ?thesis unfolding Sub by simp
qed

lemma sub_dom_neutral[simp]: fixes \<sigma> :: "('f,'v)sub" assumes "dom \<sigma> \<subseteq> V" shows "\<sigma>|_V = \<sigma>"
proof (cases \<sigma>) 
 case (Sub s) {
  fix x
  have "(if x \<in> V then s x else Var x) = (if x \<in> dom(Sub s) then s x else Var x)" using assms unfolding Sub by (cases "x \<in> dom(Sub s)") auto
 }
 hence "\<And>x.(if x \<in> V then s x else Var x) = (if x \<in> dom(Sub s) then s x else Var x)" by simp
 hence "\<And>x. (\<lambda>x. if x \<in> V then s x else Var x) x = (\<lambda>x. if x \<in> dom(Sub s) then s x else Var x) x" by simp
 hence "\<And>x. (\<lambda>x. if x \<in> V then s x else Var x) x = s x" unfolding sub_restrict_dom' by simp
 hence "(\<lambda>x. if x \<in> V then s x else Var x) = s" by simp
 hence "Sub(\<lambda>x. if x \<in> V then s x else Var x) = Sub s" by simp
 hence "(Sub s)|_V = Sub s" by simp
 thus ?thesis unfolding Sub by simp
qed

lemma coincidence_lemma': "vars_term t \<subseteq> V \<Longrightarrow> t\<cdot>(\<sigma>|_vars_term t) = t\<cdot>(\<sigma>|_V)"
proof (induct t arbitrary: V)
 case (Var x) thus ?case by (induct \<sigma>) auto
next
 case (Fun f ss)
 let ?t = "Fun f ss"
 have "\<forall>s\<in>set ss. vars_term s \<subseteq> vars_term ?t" by auto
 hence A:"\<forall>s\<in>set ss. s\<cdot>(\<sigma>|_vars_term s) = s\<cdot>(\<sigma>|_vars_term ?t)" using Fun by simp
 hence "\<forall>s\<in>set ss. vars_term s \<subseteq> V" using Fun by auto
 hence "\<forall>s\<in>set ss. s\<cdot>(\<sigma>|_(vars_term s)) = s\<cdot>(\<sigma>|_V)" using Fun by simp
 with A have "\<forall>s\<in>set ss. s\<cdot>(\<sigma>|_vars_term ?t) = s\<cdot>(\<sigma>|_V)" by auto
 thus ?case by auto
qed

lemma [simp]: "\<sigma>|_UNIV = \<sigma>" by (cases \<sigma>) auto

lemma [simp]: "\<sigma>|_{} = \<iota>" by (cases \<sigma>) auto

text {* The well-known coincidence lemma *}
lemma coincidence_lemma:"t\<cdot>\<sigma> = t\<cdot>(\<sigma>|_(vars_term t))"
proof -
 have "vars_term t \<subseteq> UNIV" by simp
 hence "t\<cdot>(\<sigma>|_vars_term t) = t\<cdot>\<sigma>" using coincidence_lemma'[where V = UNIV] by simp
 thus ?thesis by simp
qed

lemma dom_restrict:"dom(\<sigma>|_vars_term t) \<subseteq> vars_term t" by (cases \<sigma>) auto

lemma restrict_empty[simp]: assumes "x \<notin> dom \<sigma>" shows "\<sigma>|_{x} = \<iota>"
proof (cases \<sigma>)
 case (Sub s)
 have A: "\<And>x. x \<notin> dom \<sigma> \<Longrightarrow> s x = Var x" unfolding Sub by auto
 have "\<sigma>|_{x} = (Sub s)|_{x}" unfolding Sub by simp
 also have "\<dots> = Sub(\<lambda>y. if y \<in> {x} then s y else Var y)" by simp
 also have "\<dots> = Sub(\<lambda>y. if y = x then s y else Var y)" by simp
 also have "\<dots> = Sub(\<lambda>y. if y = x then s x else Var y)" by (simp cong: if_cong)
 also have "\<dots> = Sub(\<lambda>y. if y = x then Var x else Var y)" unfolding A[OF assms] by simp
 also have "\<dots> = Sub(\<lambda>y. if y = x then Var y else Var y)" by (simp cong: if_cong)
 also have "\<dots> = Sub(\<lambda>y. Var y)" by simp
 finally show ?thesis by simp
qed

lemma restrict_empty': assumes "x \<notin> dom \<sigma>" and "\<sigma>|_V = \<iota>" shows "\<sigma>|_({x} \<union> V) = \<iota>"
proof (cases \<sigma>)
 case (Sub s)
 have "Sub(\<lambda>y. if y \<in> V then s y else Var y) = Sub(\<lambda>y. Var y)" using `\<sigma>|_V = \<iota>` unfolding Sub by simp
 hence "(\<lambda>y. if y \<in> V then s y else Var y) = (\<lambda>y. Var y)" by simp
 hence A: "\<And>y. (if y \<in> V then s y else Var y) = Var y" by (rule fun_cong)
 {
  fix y
  have "(if y \<in> {x}\<union>V then s y else Var y) = Var y" 
  proof (cases "y = x")
   assume "y = x" thus ?thesis using `x \<notin> dom \<sigma>` unfolding Sub by auto
  next
   assume "y \<noteq> x" thus ?thesis using A by simp
  qed
 }
 hence "\<And>y. (if y \<in> {x}\<union>V then s y else Var y) = Var y" by simp
 thus ?thesis unfolding Sub by auto
qed

lemma restrict_empty_set:"finite V \<Longrightarrow> V \<inter> dom \<sigma> = {} \<Longrightarrow> \<sigma>|_V = \<iota>"
proof(induct rule: finite.induct)
 case emptyI thus ?case by simp
next
 case (insertI V x)
 hence "V \<inter> dom \<sigma> = {}" by simp
 with insertI have "\<sigma>|_V = \<iota>" by simp
 thus ?case using insertI restrict_empty'[where \<sigma> = \<sigma> and x = x and V = V] by simp
qed

lemma restrict_var:"x \<noteq> y \<Longrightarrow> (Var y)\<cdot>(\<sigma>|_(UNIV-{x})) = Var y\<cdot>\<sigma>" by (cases \<sigma>) auto


lemma vars_sub_ex: assumes A: "vars_term s \<subseteq> vars_term t" and B: "t\<cdot>\<sigma> = t\<cdot>\<tau>" shows "s\<cdot>\<sigma> = s\<cdot>\<tau>"
using A proof (induct s)
 case (Var x) thus ?case using B by (induct t) auto
next
 case (Fun f ss) thus ?case by auto
qed

lemma supteqp_stable:
 fixes s t :: "('f,'v)term" and \<sigma> :: "('f,'v)sub"
 assumes "s \<unrhd> t" shows "s\<cdot>\<sigma> \<unrhd> t\<cdot>\<sigma>"
using assms
proof (induct rule: supteqp.induct)
 case (refl s) show ?case by (rule refl)
next
 case (subteq_of_arg u ss t f)
 from `u \<in> set ss` have "u\<cdot>\<sigma> \<in> set (map (\<lambda>t. t\<cdot>\<sigma>) ss)" "u\<cdot>\<sigma> \<unrhd> u\<cdot>\<sigma>" by auto
 hence "(Fun f ss)\<cdot>\<sigma> \<unrhd> u\<cdot>\<sigma>" unfolding sub_apply.simps by blast
 from this and  `u\<cdot>\<sigma> \<unrhd> t\<cdot>\<sigma>` show ?case by (rule supteqp_trans)
qed

lemma suptp_stable: fixes \<sigma> :: "('f,'v)sub" assumes "s \<rhd> t" shows "s\<cdot>\<sigma> \<rhd> t\<cdot>\<sigma>"
using assms
proof (induct rule: suptp.induct)
 case (arg s ss f) 
 hence "s\<cdot>\<sigma> \<in> set (map (\<lambda>t. t\<cdot>\<sigma>) ss)" by simp
thus ?case unfolding sub_apply.simps by (rule suptp.arg)
next
 case (subt_of_arg u ss t f)
 from `u \<in> set ss` have "u\<cdot>\<sigma> \<in> set (map (\<lambda>t. t\<cdot>\<sigma>) ss)" by simp
 hence "(Fun f ss)\<cdot>\<sigma> \<rhd> u\<cdot>\<sigma>" unfolding sub_apply.simps by (rule suptp.arg)
 from this and `u\<cdot>\<sigma> \<rhd> t\<cdot>\<sigma>` show ?case by (rule suptp_trans)
qed

text {* The set of all well-formed terms w.r.t. a given signature. Variables are
arbitrary *}
inductive_set wf_fun_terms ("Term'(_')" 1000)
for F :: "('f\<times>nat)set"
where wf_var[intro]: "Var x \<in> Term(F)"
    | wf_fun[intro]: "(f,length ts) \<in> F \<Longrightarrow> (\<And>t. t \<in> set ts \<Longrightarrow> t \<in> Term(F))
        \<Longrightarrow> Fun f ts \<in> Term(F)"

fun wf_fun_terms_list :: "('f \<times> nat)list \<Rightarrow> ('f,'v)term \<Rightarrow> bool"
where "wf_fun_terms_list _ (Var _) = True"
  |  "wf_fun_terms_list sig (Fun f ts) = (
        (f,length ts) mem sig \<and> 
        forall (wf_fun_terms_list sig) ts)"

lemma wf_fun_terms_list_sound[simp]: "wf_fun_terms_list sig t = (t \<in> Term(set sig))"
proof (induct t, simp add: wf_var)
  case (Fun f ts)
  show ?case (is "?l = ?r")
  proof
    assume ?l
    with Fun have "\<forall> t \<in> set ts. t \<in> Term(set sig)" and "(f,length ts) \<in> set sig" by (auto simp: mem_iff)
    thus ?r by auto
  next
    assume ?r
    thus ?l
    proof (cases rule: wf_fun_terms.cases, simp)
      case (wf_fun f' ts')
      with Fun show ?thesis by (auto simp: mem_iff)
    qed
  qed
qed

declare wf_fun_terms_list.simps[simp del]

fun clean :: "(('f,'v)term \<Rightarrow> 'v) \<Rightarrow> ('f\<times>nat)set \<Rightarrow> ('f,'v)term \<Rightarrow> ('f,'v)term"
where "clean v F (Var y) = Var y"
    | "clean v F (Fun f ts) = (
        if (f,length ts) \<in> F then Fun f (map (clean v F) ts)
                             else Var(v(Fun f ts)))"

lemma clean_wf_fun_terms[simp]: "clean v F t \<in> Term(F)"
proof (induct t)
  case (Var x) show ?case by auto
next
  case (Fun f ts)
  hence IH: "\<And>s. s \<in> set ts \<Longrightarrow> clean v F s \<in> Term(F)" by simp
  thus ?case
  proof (cases "(f,length ts) \<in> F")
    case True with IH show ?thesis by force
  next
    case False thus ?thesis by auto
  qed
qed

lemma clean_id[simp]: "clean v F (clean v F t) = clean v F t" by (induct t) (auto)

lemma clean_sub_apply[simp]:
  assumes "t \<in> Term(F)"
  shows "clean v F (t\<cdot>(Sub s)) = t\<cdot>(Sub(clean v F o s))"
using assms proof (induct t rule: term_induct)
  case (Var x) show ?case by simp
next
  case (Fun f ss)
  from `Fun f ss \<in> Term(F)`
    have "\<forall>t\<in>set ss. t \<in> Term(F)" by (cases rule: wf_fun_terms.cases) simp_all
  with Fun have IH: "\<And>t. t \<in> set ss \<Longrightarrow> clean v F (t \<cdot> Sub s) = t \<cdot> Sub(clean v F o s)" by best
  show ?case
  proof (cases "(f,length ss) \<in> F")
    case True show ?thesis by (auto simp: o_def IH True)
  next
    case False with `Fun f ss \<in> Term(F)` have False by cases auto
    thus ?thesis by simp
  qed
qed

lemma wf_fun_terms_subset[simp]: assumes "F \<subseteq> G" shows "Term(F) \<subseteq> Term(G)"
proof
  fix t assume "t \<in> Term(F)" thus "t \<in> Term(G)" using assms by (induct t) auto
qed

lemma funas_term_id[simp]: fixes t::"('f,'v)term" shows "t \<in> Term(funas_term t)"
proof (induct t)
  case (Var x) show ?case by auto
next
  case (Fun f ss)
  have "\<forall>s\<in>set ss. funas_term s \<subseteq> funas_term(Fun f ss)" by auto
  with Fun have IH: "\<And>s. s \<in> set ss \<Longrightarrow> s \<in> Term(funas_term(Fun f ss))"
    using wf_fun_terms_subset[where G="funas_term(Fun f ss)"] by blast
  have "(f,length ss) \<in> funas_term(Fun f ss)" by simp
  with IH show ?case by fast
qed

fun funas_ctxt :: "('f,'v)ctxt \<Rightarrow> ('f\<times>nat)set"
where "funas_ctxt Hole = {}"
    | "funas_ctxt(More f ss1 D ss2) = {(f,Suc(length(ss1@ss2)))} \<union>
        funas_ctxt D \<union> \<Union>(set(map funas_term (ss1@ss2)))"

lemma funas_term_ctxt_apply[simp]:"funas_term(C\<langle>t\<rangle>) = funas_ctxt C \<union> funas_term t"
proof (induct t)
  case (Var x) show ?case by (induct C) auto
next
  case (Fun f ts) show ?case by (induct C arbitrary: f ts) auto
qed

lemma funas_term_sub_apply: "funas_term(t\<cdot>\<sigma>) = funas_term t \<union> (\<Union>x\<in>vars_term t. funas_term(Var x \<cdot> \<sigma>))" by (induct t) auto


inductive_set wf_fun_ctxts ("Ctxt'(_')" 1000)
for F::"('f\<times>nat)set"
where hole[intro]: "Hole \<in> Ctxt(F)"
    | more[intro]: "(f,Suc(length(ss1@ss2))) \<in> F \<Longrightarrow>
                    (\<And>t. t \<in> set(ss1@ss2) \<Longrightarrow> t \<in> Term(F)) \<Longrightarrow>
                    D \<in> Ctxt(F) \<Longrightarrow> More f ss1 D ss2 \<in> Ctxt(F)"

lemma supteqp_Var_id[simp]: assumes "Var x \<unrhd> s" shows "s = Var x"
using assms proof (induct s rule: term_induct)
  case (Var y) thus ?case by (cases rule: supteqp.cases) simp_all
next
  case (Fun f ts)
  from `Var x \<unrhd> Fun f ts` have False by (cases rule: supteqp.cases) auto
  thus ?case by simp
qed

lemma supteqp_imp_wf_fun_terms: assumes "t \<in> Term(F)" and "t \<unrhd> s" shows "s \<in> Term(F)"
using assms proof (induct t)
  case (wf_var x) from supteqp_Var_id[OF wf_var] show ?case by auto
next
  case (wf_fun f ts)
  from `Fun f ts \<unrhd> s` have "s = Fun f ts \<or> (\<exists>u\<in>set ts. u \<unrhd> s)" by (cases rule: supteqp.cases) auto
  thus ?case proof
    assume "s = Fun f ts" thus ?thesis using wf_fun wf_fun_terms.wf_fun by auto
  next
    assume "\<exists>u\<in>set ts. u \<unrhd> s"
    then obtain u where "u \<in> set ts" and "u \<unrhd> s" by auto
    with wf_fun show ?thesis by auto
  qed
qed

lemma [simp]: assumes "t \<in> set ts" shows "Fun f ts \<noteq> t"
proof (rule ccontr)
  assume "\<not> Fun f ts \<noteq>  t"
  hence "size(Fun f ts) = size t" by simp
  moreover from assms have "size t < list_size size ts" by (induct ts) auto
  ultimately show False by simp
qed

lemma [simp]: assumes "t \<in> set ts" shows "Fun f ts \<unrhd> t"
using assms by auto
  
fun args :: "('f,'v)term \<Rightarrow> ('f,'v)term list"
where "args(Var _) = []"
    | "args(Fun _ ts) = ts"

lemma suptp_imp_args: assumes "\<forall>t. s \<rhd> t \<longrightarrow> P t" shows "\<forall>t\<in>set(args s). P t"
using assms by (cases s) simp_all

lemma ctxt_apply_eq_False[simp]: assumes "t = (More f ss1 D ss2)\<langle>t\<rangle>" shows "False"
proof -
  from assms and nectxt_imp_suptp have "t \<rhd> t" by best
  thus "False" by simp
qed

lemma wf_fun_terms_Fun[simp]: assumes "Fun f ts \<in> Term(F)" shows "(f,length ts) \<in> F"
using assms by cases auto

lemma clean_fun_terms_apply_id[simp]:
  assumes "t \<in> Term(F)" shows "clean v F t = t"
using assms proof (induct t rule: term_induct)
  case (Var x) show ?case by simp
next
  case (Fun f ts)
  hence IH: "\<And>s. s \<in> set ts \<Longrightarrow> clean v F s = s" by (auto simp: supteqp_imp_wf_fun_terms)
  thus ?case
  proof (cases "(f,length ts) \<in> F")
    case False with Fun show ?thesis by simp
  next
    case True thus ?thesis using map_idI[OF IH,of ts] by auto
  qed
qed

lemma supteqp_imp_funas_subset[simp]: "s \<unrhd> t \<Longrightarrow> funas_term t \<subseteq> funas_term s" by (induct rule: supteqp.induct) auto

lemma ctxt_clean_sub_wf: "C \<in> Ctxt(F) \<Longrightarrow> C \<cdot>c (Sub(clean v F \<circ> \<sigma>)) \<in> Ctxt(F)"
proof (induct C)
  case Hole show ?case by auto
next
  case (More f ss1 E ss2)
  let ?C = "More f ss1 E ss2"
  from `?C \<in> Ctxt(F)` have "(f,Suc(length(ss1@ss2))) \<in> F \<and> (\<forall>t\<in>set(ss1@ss2). t \<in> Term(F))" by cases auto
  hence in_F: "(f,Suc(length(ss1@ss2))) \<in> F" and all: "\<forall>t\<in>set(ss1@ss2). t \<in> Term(F)" by auto
  from in_F have in_F': "(f,Suc(length((map (\<lambda>t. t\<cdot>Sub(clean v F \<circ> \<sigma>)) ss1)@(map (\<lambda>t. t\<cdot>Sub(clean v F \<circ> \<sigma>)) ss2)))) \<in> F" by simp
  from all have all': "\<forall>t\<in>set(ss1@ss2). t\<cdot>Sub(clean v F \<circ> \<sigma>) = clean v F (t\<cdot>Sub \<sigma>)" by auto
  have "\<forall>t\<in>set(ss1@ss2). clean v F (t\<cdot>Sub \<sigma>) \<in> Term(F)" by auto
  with all' have "\<forall>t\<in>set(ss1@ss2). t\<cdot>Sub(clean v F \<circ> \<sigma>) \<in> Term(F)" by auto
  hence "\<forall>t\<in>set(map (\<lambda>t. t\<cdot>Sub(clean v F \<circ> \<sigma>)) (ss1@ss2)). t \<in> Term(F)" by auto
  hence args: "\<And>t. t \<in> set(map (\<lambda>t. t\<cdot>Sub(clean v F \<circ> \<sigma>)) ss1 @ map (\<lambda>t. t\<cdot>Sub(clean v F \<circ> \<sigma>)) ss2) \<Longrightarrow> t \<in> Term(F)" by auto
  from `?C \<in> Ctxt(F)` have "E \<in> Ctxt(F)" by cases auto
  with More have IH: "E \<cdot>c Sub(clean v F \<circ> \<sigma>) \<in> Ctxt(F)" by best
  from args wf_fun_ctxts.more[OF in_F' _ this] show ?case unfolding sub_apply_to_ctxt.simps by best
qed

lemma ctxt_compose_wf[simp]: "C \<in> Ctxt(F) \<Longrightarrow> E \<in> Ctxt(F) \<Longrightarrow> (C \<circ>c E) \<in> Ctxt(F)"
proof (induct C)
  case Hole thus ?case by simp
next
  case (More f ss1 G ss2)
  let ?C = "More f ss1 G ss2"
  from `?C \<in> Ctxt(F)` have G: "G \<in> Ctxt(F)" by cases auto
  from `?C \<in> Ctxt(F)` have "(f,Suc(length(ss1@ss2))) \<in> F \<and> (\<forall>t\<in>set(ss1@ss2). t \<in> Term(F))" by cases auto
  hence in_F: "(f,Suc(length(ss1@ss2))) \<in> F" and args: "\<And>t. t \<in> set(ss1@ss2) \<Longrightarrow> t \<in> Term(F)" by auto
  from G More have IH: "(G \<circ>c E) \<in> Ctxt(F)" by simp
  from wf_fun_ctxts.more[OF in_F _ IH] args show ?case by auto
qed

lemma set_set_cons: assumes "P x" and set:"\<And>y. y \<in> set xs \<Longrightarrow> P y" shows "y \<in> set(x#xs) \<Longrightarrow> P y"
using assms by auto

lemma ctxt_apply_wf[simp]: 
  fixes C::"('f,'v)ctxt" and t::"('f,'v)term" shows "C \<in> Ctxt(F) \<Longrightarrow> t \<in> Term(F) \<Longrightarrow> C\<langle>t\<rangle> \<in> Term(F)"
proof (induct C)
  case Hole thus ?case by simp
next
  case (More f ss1 E ss2)
  let ?C = "(More f ss1 E ss2::('f,'v)ctxt)"
  from `?C \<in> Ctxt(F)` have E: "E \<in> Ctxt(F)" by cases auto
  with More have IH: "E\<langle>t\<rangle> \<in> Term(F)" by simp
  from `?C \<in> Ctxt(F)` have "(f,Suc(length(ss1@ss2))) \<in> F \<and> (\<forall>t\<in>set(ss1@ss2). t \<in> Term(F))" by cases auto
  hence in_F: "(f,Suc(length(ss1@ss2))) \<in> F" and args: "\<And>(t::('f,'v)term). t \<in> set(ss1@ss2) \<Longrightarrow> t \<in> Term(F)" by auto
  have len: "length(ss1@E\<langle>t\<rangle>#ss2) = Suc(length(ss1@ss2))" by auto
  from in_F have in_F': "(f,length(ss1@E\<langle>t\<rangle>#ss2)) \<in> F" unfolding len .
  from args have pre: "\<forall>(t::('f,'v)term) \<in> set ss1. t \<in> Term(F)" by simp
  from args have "\<And>(t::('f,'v)term). t \<in> set ss2 \<Longrightarrow> t \<in> Term(F)" by simp
  with set_set_cons[of "\<lambda>(t::('f,'v)term). t \<in> Term(F)" "E\<langle>t\<rangle>" ss2,OF IH this] have "\<forall>u \<in> set(E\<langle>t\<rangle>#ss2). u \<in> Term(F)" by blast
  with pre have a: "\<forall>(u::('f,'v)term) \<in> set(ss1@E\<langle>t\<rangle>#ss2). u \<in> Term(F)" by auto
  have "\<And>(u::('f,'v)term). u \<in> set(ss1@E\<langle>t\<rangle>#ss2) \<Longrightarrow> u \<in> Term(F)" using a by best
  with wf_fun_terms.wf_fun[OF in_F'] show ?case by auto
qed

lemma clean_sub_apply_wf[simp]: assumes "t \<in> Term(F)" shows "t\<cdot>Sub(clean v F \<circ> s) \<in> Term(F)"
proof -
  from assms have clean: "clean v F (t\<cdot>Sub s) = t\<cdot>Sub(clean v F \<circ> s)" by auto
  have "clean v F (t\<cdot>Sub s) \<in> Term(F)" by simp
  thus ?thesis unfolding clean .
qed

lemma ctxt_apply_wf_imp_ctxt_and_term_wf: assumes "C\<langle>t\<rangle> \<in> Term(F)" shows "C \<in> Ctxt(F) \<and> t \<in> Term(F)"
using assms proof (induct C)
  case Hole thus ?case by auto
next
  case (More f ss1 E ss2)
  let ?C = "More f ss1 E ss2"
  from `?C\<langle>t\<rangle> \<in> Term(F)` have "(f,length(ss1@E\<langle>t\<rangle>#ss2)) \<in> F \<and> (\<forall>s\<in>set(ss1@E\<langle>t\<rangle>#ss2). s \<in> Term(F))"
    unfolding ctxt_apply.simps by cases auto
  hence in_F: "(f,length(ss1@E\<langle>t\<rangle>#ss2)) \<in> F" and args: "\<forall>s\<in>set(ss1@E\<langle>t\<rangle>#ss2). s \<in> Term(F)" by auto
  from in_F have in_F': "(f,Suc(length(ss1@ss2))) \<in> F" by auto
  from args have "E\<langle>t\<rangle> \<in> Term(F)" by simp
  with More have "E \<in> Ctxt(F) \<and> t \<in> Term(F)" by best
  hence "E \<in> Ctxt(F)" and "t \<in> Term(F)" by auto
  from args have "\<forall>s\<in>set(ss1). s \<in> Term(F)" by auto
  moreover from args have "\<forall>s\<in>set(ss2). s \<in> Term(F)" by auto
  ultimately have "\<forall>s\<in>set(ss1@ss2). s \<in> Term(F)" by auto
  hence "\<And>s. s \<in> set(ss1@ss2) \<Longrightarrow> s \<in> Term(F)" by auto
  from wf_fun_ctxts.more[OF in_F' _ `E \<in> Ctxt(F)`] this `t \<in> Term(F)` show ?case by auto
qed

lemma clean_ctxt_apply: assumes "C \<in> Ctxt(F)" shows "clean v F (C\<langle>t\<rangle>) = C\<langle>clean v F t\<rangle>"
using assms proof (induct C)
  case hole thus ?case by simp
next
  case (more f ss1 ss2 E)
  let ?C = "More f ss1 E ss2"
  from more have "(f,Suc(length(ss1@ss2))) \<in> F" by auto
  hence in_F: "(f,length(ss1@E\<langle>clean v F t\<rangle>#ss2)) \<in> F" by auto
  from more have "clean v F (?C\<langle>t\<rangle>) = Fun f (map (clean v F) (ss1@E\<langle>t\<rangle>#ss2))" by auto
  from `\<And>t. t \<in> set(ss1@ss2) \<Longrightarrow> t \<in> Term(F)` have args: "\<forall>t\<in>set(ss1@ss2). clean v F t = t" by auto
  hence "\<forall>t\<in>set ss1. clean v F t = t" by auto
  hence pre: "map (clean v F) ss1 = ss1" using map_idI[of ss1 "clean v F"] by best
  from args have "\<forall>t\<in>set ss2. clean v F t = t" by auto
  hence "map (clean v F) ss2 = ss2" using map_idI[of ss2 "clean v F"] by best
  with `clean v F (E\<langle>t\<rangle>) = E\<langle>clean v F t\<rangle>` have post: "map (clean v F) (E\<langle>t\<rangle>#ss2) = E\<langle>clean v F t\<rangle>#ss2" by auto
  have "map (clean v F) (ss1@E\<langle>t\<rangle>#ss2) = map (clean v F) ss1 @ map (clean v F) (E\<langle>t\<rangle>#ss2)" by simp
  also have "\<dots> = ss1@E\<langle>clean v F t\<rangle>#ss2" unfolding pre post by simp
  finally have "map (clean v F) (ss1@E\<langle>t\<rangle>#ss2) = ss1@E\<langle>clean v F t\<rangle>#ss2" .
  with in_F show ?case by simp
qed

lemma ctxt_power_compose_distr: "C ^ (m + n) = C ^ m \<circ>c C ^ n" by (induct m) (simp_all add: one_ctxt_def times_ctxt_def)

fun vars_ctxt :: "('f,'v)ctxt \<Rightarrow> 'v set"
where "vars_ctxt \<box> = {}"
    | "vars_ctxt(More f ss1 D ss2) = vars_ctxt D \<union> \<Union>set(map vars_term (ss1@ss2))"

lemma sub_apply_id': assumes "vars_term t \<inter> V = {}" shows "t\<cdot>(\<sigma>|_V) = t"
using assms proof (induct t)
  case (Var x) thus ?case by (cases \<sigma>) simp_all
next
  case (Fun f ts)
  hence IH: "\<forall>s\<in>set ts. s\<cdot>(\<sigma>|_V) = s" by auto
  with map_idI[of ts "\<lambda>t. t\<cdot>(\<sigma>|_V)"] show ?case by simp
qed

lemma sub_apply_ctxt_id: assumes "vars_ctxt C \<inter> V = {}" shows "C \<cdot>c (\<sigma>|_V) = C"
using assms proof (induct C)
  case Hole show ?case by simp
next
  case (More f ss1 D ss2)
  hence IH: "D \<cdot>c (\<sigma>|_V) = D" by auto
  from More have "\<forall>s\<in>set(ss1@ss2). vars_term s \<inter> V = {}" by auto
  with sub_apply_id' have args: "\<forall>s\<in>set(ss1@ss2). s\<cdot>(\<sigma>|_V) = s" by best
  from args have "\<forall>s\<in>set ss1. s\<cdot>(\<sigma>|_V) = s" by simp
  with map_idI[of ss1 "\<lambda>t. t\<cdot>(\<sigma>|_V)"] have ss1: "map (\<lambda>s. s\<cdot>(\<sigma>|_V)) ss1 = ss1" by best
  from args have "\<forall>s\<in>set ss2. s\<cdot>(\<sigma>|_V) = s" by simp
  with map_idI[of ss2 "\<lambda>t. t\<cdot>(\<sigma>|_V)"] have ss2: "map (\<lambda>s. s\<cdot>(\<sigma>|_V)) ss2 = ss2" by best
  show ?case by (simp add: ss1 ss2 IH)
qed

lemma vars_term_Var_id: "vars_term o Var = (\<lambda>x. {x})" by (rule ext) simp

lemma ctxt_exhaust_rev[case_names Hole More]:
  assumes "C = \<box> \<Longrightarrow> P" and "\<And>D f ss1 ss2. C = D \<circ>c (More f ss1 \<box> ss2) \<Longrightarrow> P" shows "P"
proof (cases C)
  case Hole with assms show ?thesis by simp
next
  case (More g ts1 E ts2)
  hence "\<exists>D f ss1 ss2. C = D \<circ>c (More f ss1 \<box> ss2)"
  proof (induct E arbitrary: C g ts1 ts2)
    case Hole hence "C = \<box> \<circ>c (More g ts1 \<box> ts2)" by simp
    thus ?case by best
  next
    case (More h us1 F us2)
    from More(1)[of "More h us1 F us2"]
     obtain G i vs1 vs2 where IH: "More h us1 F us2 = G \<circ>c More i vs1 \<box> vs2" by force
    from More have "C = (More g ts1 \<box> ts2 \<circ>c G) \<circ>c More i vs1 \<box> vs2" unfolding IH by simp
    thus ?case by best
  qed
  thus ?thesis using assms by auto
qed

lemma ctxt_compose_wf_imp_ctxts_wf:
  assumes "C \<circ>c D \<in> Ctxt(F)" shows "C \<in> Ctxt(F) \<and> D \<in> Ctxt(F)"
using assms proof (induct C)
  case Hole thus ?case by auto
next
  case (More f ss1 E ss2)
  from More(2) have wf: "(f,Suc(length(ss1@ss2))) \<in> F \<and> (\<forall>t\<in>set(ss1@ss2). t \<in> Term(F)) \<and> E \<circ>c D \<in> Ctxt(F)"
    unfolding ctxt_compose.simps by (cases rule:wf_fun_ctxts.cases) auto
  with More have IH: "E \<in> Ctxt(F) \<and> D \<in> Ctxt(F)" by fast
  with wf show ?case by auto
qed

fun sub_extend :: "('f,'v)sub \<Rightarrow> ('v\<times>('f,'v)term)list \<Rightarrow> ('f,'v)sub"
where "sub_extend (Sub s) vts = Sub(\<lambda>x. case lookup x vts of Some t \<Rightarrow> t
                                                           | None   \<Rightarrow> s(x))"

lemma lookup_None[simp]: assumes "V \<inter> set vs = {}" and "x \<in> V" shows "lookup x (zip vs ts) = None"
using assms proof (induct vs arbitrary: ts)
  case Nil show ?case by simp
next
  case (Cons v vs)
  note IH = this
  hence neq: "v \<noteq> x" by auto
  show ?case
  proof (cases ts)
    case Nil show ?thesis unfolding Nil by simp
  next
     case (Cons t ts')
     from IH have IH': "lookup x (zip vs ts') = None" by auto
     have "lookup x (zip (v#vs) ts) = lookup x ((v,t)#zip vs ts')" unfolding Cons by simp
     also have "\<dots> = lookup x (zip vs ts')" using neq by simp
     also have "\<dots> = None" unfolding IH' by simp
     finally show ?thesis .
  qed
qed

lemma sub_extend_id:
  assumes "V \<inter> set vs = {}" and "vars_term t \<subseteq> V"
  shows "t \<cdot> sub_extend \<sigma> (zip vs ts) = t \<cdot> \<sigma>"
using assms by (induct t,cases \<sigma>) auto

lemma lookup_Some[simp]:
  assumes "distinct vs" and "length vs \<le> length ts" and "i < length vs"
  shows "lookup (vs!i) (zip vs ts) = Some(ts!i)"
using assms proof (induct i arbitrary: vs ts)
  case 0 thus ?case by (induct ts) (induct vs,auto)+
next
  case (Suc i) show ?case
  proof (cases ts)
    case Nil with Suc show ?thesis unfolding Nil by simp
  next
    case (Cons t ts')
    note Cons' = this
    show ?thesis
    proof (cases vs)
      case Nil from Suc show ?thesis unfolding Cons' Nil by simp
    next
      case (Cons v vs') from Suc show ?thesis unfolding Cons Cons' by auto
    qed
  qed
qed

lemma sub_extend_absorb:
  assumes "distinct vs" and "length vs = length ss"
  shows "map (\<lambda>t. t\<cdot>(sub_extend \<sigma> (zip vs ss))) (map Var vs) = ss" (is "?ss = _")
proof -
  obtain s where \<sigma>: "\<sigma> = Sub s" by (cases \<sigma>) auto
  let ?\<sigma> = "sub_extend (Sub s) (zip vs ss)"
  from assms have "length vs \<le> length ss" by simp
  from assms have "length ?ss = length ss" by simp
  moreover have "\<forall>i<length ?ss. ?ss!i = ss!i"
  proof (intro allI impI)
    fix i assume "i < length ?ss"
    hence i: "i < length(map Var vs)" by simp
    hence len: "i < length vs" by simp
    have "?ss!i = (map Var vs!i)\<cdot>?\<sigma>" unfolding nth_map[OF i,of "\<lambda>t. t\<cdot>?\<sigma>"] \<sigma> by simp
    also have "\<dots> = Var(vs!i)\<cdot>?\<sigma>" unfolding nth_map[OF len] by simp
    also have "\<dots> = (case lookup (vs!i) (zip vs ss) of None \<Rightarrow> s(vs!i) | Some t \<Rightarrow> t)" by simp
    also have "\<dots> = ss!i" unfolding lookup_Some[OF `distinct vs` `length vs \<le> length ss` len] by simp
    finally show "?ss!i = ss!i" by simp
  qed
  ultimately show ?thesis by (rule nth_equalityI)
qed

lemma vars_term_finite[simp]:
  fixes t::"('f,'v)term" shows "finite(vars_term t)" by (induct t) simp_all

fun linear_term :: "('f,'v)term \<Rightarrow> bool"
where [code del]: "linear_term(Var _) = True"
    | "linear_term(Fun _ ts) = (is_partition(map vars_term ts) \<and> forall linear_term ts)"

end
